#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #
import json
import re
from datetime import datetime, timedelta

from .events import ControlEvent
from ..bridge import BridgeContext
from ...utils import trans


class Voice:
    def __init__(self, window=None):
        """
        Voice core

        :param window: Window instance
        """
        self.window = window
        self.commands = {
            ControlEvent.APP_STATUS: "Get the current application status",
            ControlEvent.APP_EXIT: "Exit the application",
            ControlEvent.AUDIO_OUTPUT_ENABLE: "Enable audio output",
            ControlEvent.AUDIO_OUTPUT_DISABLE: "Disable audio output",
            ControlEvent.AUDIO_INPUT_ENABLE: "Enable audio input",
            ControlEvent.AUDIO_INPUT_DISABLE: "Disable audio input",
            ControlEvent.CALENDAR_ADD: "Add a memo to the calendar",
            ControlEvent.CALENDAR_CLEAR: "Clear memos from calendar",
            ControlEvent.CALENDAR_READ: "Read the calendar memos",
            ControlEvent.CAMERA_ENABLE: "Enable the camera",
            ControlEvent.CAMERA_DISABLE: "Disable the camera",
            ControlEvent.CAMERA_CAPTURE: "Capture the camera",
            ControlEvent.CMD_CONFIRM: "Confirmation of the command",
            ControlEvent.CTX_NEW: "Create a new context",
            ControlEvent.CTX_PREV: "Go to the previous context",
            ControlEvent.CTX_NEXT: "Go to the next context",
            ControlEvent.CTX_LAST: "Go to the last context",
            ControlEvent.CTX_INPUT_FOCUS: "Focus on the input",
            ControlEvent.CTX_INPUT_SEND: "Send the input",
            ControlEvent.CTX_INPUT_CLEAR: "Clear the input",
            ControlEvent.CTX_CURRENT: "Get current conversation info",
            ControlEvent.CTX_STOP: "Stop executing current action",
            ControlEvent.CTX_ATTACHMENTS_CLEAR: "Clear the attachments",
            ControlEvent.CTX_READ_LAST: "Read the last conversation entry",
            ControlEvent.CTX_READ_ALL: "Read the whole conversation",
            ControlEvent.CTX_RENAME: "Rename current context",
            ControlEvent.CTX_SEARCH_STRING: "Search for a string in the conversation",
            ControlEvent.CTX_SEARCH_CLEAR: "Clear the search results",
            ControlEvent.INPUT_SEND: "Send the message to input",
            ControlEvent.INPUT_APPEND: "Append message to current input without sending it",
            ControlEvent.MODE_CHAT: "Switch to chat mode",
            ControlEvent.MODE_LLAMA_INDEX: "Switch to Chat with files (llama-index) mode",
            ControlEvent.NOTE_ADD: "Add note to notepad",
            ControlEvent.NOTEPAD_CLEAR: "Clear notepad contents",
            ControlEvent.NOTEPAD_READ: "Read current notepad contents",
            ControlEvent.TAB_CHAT: "Switch to the chat tab",
            ControlEvent.TAB_CALENDAR: "Switch to the calendar tab",
            ControlEvent.TAB_DRAW: "Switch to the draw (painter) tab",
            ControlEvent.TAB_FILES: "Switch to the files tab",
            ControlEvent.TAB_NOTEPAD: "Switch to the notepad tab",
            ControlEvent.TAB_NEXT: "Switch to the next tab",
            ControlEvent.TAB_PREV: "Switch to the previous tab",
            ControlEvent.VOICE_MESSAGE_START: "Start listening for voice input",
            ControlEvent.VOICE_MESSAGE_STOP: "Stop listening for voice input",
            ControlEvent.VOICE_MESSAGE_TOGGLE: "Toggle listening for voice input",
        }

    def get_prompt(self, text: str):
        """
        Get prompt for voice

        :param text: text to be spoken
        :return: prompt
        """
        commands_str = "\n".join([f"{k} = {v}" for k, v in self.commands.items()])
        prompt = """
        Recognize the voice command and select the corresponding command from the list below. 
        Return the chosen command ID as a JSON string in the following syntax:

        {"cmd": "command_id", params: "optional message"}
        
        If no command matches the voice request, then return "unknown" as the command ID.
        If user provide additional message, it should be extracted from voice command and included (WITHOUT the command part) in the JSON response in "params" param.
        
        Important: Only the JSON specified above should be returned in the response, without any additional text.
        
        Available command IDs (with descriptions):
        ----------------------
        """+commands_str+"""
        
        
        User's voice input to recognize:
        
        """+text

        return prompt

    def extract_json(self, text: str):
        """
        Extract JSON from text

        :param text: text
        :return: JSON
        """
        json_pattern = r'(\{.*?\})'
        cmds = []
        matches = re.findall(json_pattern, text, re.DOTALL)
        if len(matches) > 0:
            for match in matches:
                try:
                    data = json.loads(match)
                    if "cmd" in data:
                        cmd = data["cmd"]
                        if cmd in self.commands:
                            item = {
                                "cmd": cmd,
                                "params": data.get("params", "")
                            }
                            cmds.append(item)
                except Exception as e:
                    self.window.core.debug.log(e)
        return cmds

    def recognize_commands(self, text: str) -> list:
        """
        Recognize voice command

        :param text: voice input
        :return: recognized command
        """
        prompt = self.get_prompt(text)
        model = self.window.core.models.from_defaults()
        tmp_model = self.window.core.config.get("access.voice_control.model", "gpt-3.5-turbo")
        if self.window.core.models.has(tmp_model):
            model = self.window.core.models.get(tmp_model)
        bridge_context = BridgeContext(
            prompt=prompt,
            system_prompt="You are a helpful assistant",
            model=model,  # model instance
            max_tokens=0,
            temperature=0.0,
        )
        response = self.window.core.bridge.quick_call(
            context=bridge_context,
        )
        if response is None or response == "":
            return []

        return self.extract_json(response)

    def get_status(self):
        """
        Get the current application status

        :return: status, text to read
        """
        ctx = ""
        total = self.window.core.ctx.count_meta()
        meta = self.window.core.ctx.get_current_meta()
        mode = trans("mode." + self.window.core.config.get("mode"))
        tab = self.window.controller.ui.get_current_tab_name()
        last = ""
        msg = ""
        if meta is not None:
            ctx = meta.name
            last = self.convert_date(meta.updated)
        try:
            msg = trans("event.audio.app.status").format(
                mode=mode,
                tab=tab,
                ctx=ctx,
                last=last,
                total=total,
            )
        except Exception as e:
            pass
        return msg

    def get_current_ctx(self):
        """
        Get current context

        :return: text to read
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is not None:
            name = meta.name
            updated = self.convert_date(meta.updated)
            try:
                msg = trans("event.audio.ctx.current").format(
                    ctx=name,
                    last=updated,
                )
            except Exception as e:
                msg = ""
            return msg
        return ""

    def get_selected_ctx(self):
        """
        Get current context

        :return: text to read
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is not None:
            name = meta.name
            updated = self.convert_date(meta.updated)
            try:
                msg = trans("event.audio.ctx.selected").format(
                    ctx=name,
                    last=updated,
                )
            except Exception as e:
                msg = ""
            return msg
        return ""

    def get_selected_tab(self):
        """
        Get current tab

        :return: text to read
        """
        msg = ""
        tab = self.window.controller.ui.get_current_tab_name()
        try:
            msg = trans("event.audio.tab.switch").format(
                tab=tab,
            )
        except Exception as e:
            pass
        return msg

    def get_last_ctx_item(self):
        """
        Get last context item

        :return: text to read
        """
        text = ""
        item = self.window.core.ctx.get_last()
        if item is not None:
            try:
                text = trans("event.audio.ctx.last").format(
                    input=item.input,
                    output=item.output,
                )
            except Exception as e:
                pass
        return text

    def get_all_ctx_items(self) -> str:
        """
        Get all context items

        :return: text to read
        """
        text = ""
        entries = []
        items = self.window.core.ctx.get_all_items(ignore_first=False)
        for item in items:
            try:
                entries.append(trans("event.audio.ctx.last").format(
                    input=item.input,
                    output=item.output,
                ))
            except Exception as e:
                pass
        if len(entries) > 0:
            text = "\n".join(entries)
        return text

    def convert_date(self, timestamp: int) -> str:
        """
        Convert timestamp to human readable format

        :param timestamp: timestamp
        :return: string
        """
        today = datetime.today().date()
        yesterday = today - timedelta(days=1)
        date = datetime.fromtimestamp(timestamp).date()
        hour_min = datetime.fromtimestamp(timestamp).strftime("%H:%M")

        days_ago = (today - date).days
        weeks_ago = days_ago // 7

        if date == today:
            return trans('dt.today') + " " + hour_min
        elif date == yesterday:
            return trans('dt.yesterday') + " " + hour_min
        elif weeks_ago == 1:
            return trans('dt.week')
        elif 1 < weeks_ago < 4:
            return f"{weeks_ago} " + trans('dt.weeks')
        elif days_ago < 30:
            return f"{days_ago} " + trans('dt.days_ago')
        elif 30 <= days_ago < 32:
            return trans('dt.month')
        else:
            return date.strftime("%Y-%m-%d")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.16 12:00:00                  #
# ================================================== #

import json
from urllib.parse import quote

from PySide6.QtCore import Slot

from pygpt_net.plugin.base import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = BaseSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        msg = None
        for item in self.cmds:
            for my_cmd in self.plugin.get_option_value("cmds"):
                if my_cmd["name"] == item["cmd"]:
                    try:
                        response = self.handle_cmd(my_cmd, item)
                        if response is False:
                            continue
                        responses.append(response)

                    except Exception as e:
                        msg = "Error: {}".format(e)
                        responses.append({
                            "request": {
                                "cmd": item["cmd"],
                            },
                            "result": "Error {}".format(e),
                        })
                        self.error(e)
                        self.log(msg)

        # send response
        if len(responses) > 0:
            for response in responses:
                self.reply(response)

        # update status
        if msg is not None:
            self.status(msg)

    def handle_cmd(self, command: dict, item: dict) -> dict or bool:
        """
        Handle API command

        :param command: command configuration
        :param item: requested item
        :return: response or False
        """
        request = {
            "cmd": command["name"],
            "type": command["type"],
            "url": command["endpoint"],
        }
        post_params = {}
        post_json_tpl = ""
        headers = {}  # HTTP headers, if any, Auth, API key, etc.
        result = None
        if 'post_json' in command and command["post_json"].strip() != "":
            post_json_tpl = str(command["post_json"])  # POST JSON, as string

        # prepare call
        endpoint = command["endpoint"]  # API endpoint URL

        # extract extra headers from JSON if any
        if 'headers' in command and command["headers"].strip() != "":
            try:
                headers = json.loads(command["headers"])  # key-value pairs
            except json.JSONDecodeError as e:
                msg = "Error decoding headers JSON: {}".format(e)
                self.log(msg)
                return False # abort

        # append GET params to endpoint URL placeholders
        if 'get_params' in command and command["get_params"].strip() != "":
            # append params to cmd placeholders
            params_list = self.plugin.extract_params(
                command["get_params"],
            )
            for p in params_list:
                param = p["name"]
                if param in item["params"]:
                    endpoint = endpoint.replace(
                        "{" + param + "}",  # replace { placeholder }
                        quote(str(item["params"][param])),
                    )

        # append POST params to POST data placeholders and POST JSON
        if 'post_params' in command and command["post_params"].strip() != "":
            # append params to post params
            params_list = self.plugin.extract_params(
                command["post_params"],
            )
            for p in params_list:
                param = p["name"]
                if param in item["params"]:
                    post_params[param] = item["params"][param]

                    # append to POST JSON
                    post_json_tpl = post_json_tpl.replace(
                        "%" + param + "%",  # replace % placeholder
                        str(item["params"][param]),
                    )

        # check if endpoint is not empty
        if endpoint is None or endpoint == "":
            msg = "Endpoint URL is empty!"
            self.log(msg)
            return False # abort

        # check if type is not empty
        if command["type"] not in ["POST", "POST_JSON", "GET"]:
            command["type"] = "GET"

        # POST
        if command["type"] == "POST":
            msg = "[POST] Calling API endpoint: {}".format(endpoint)
            self.log(msg)
            result = self.plugin.call_post(endpoint, post_params, headers)

        # POST JSON
        elif command["type"] == "POST_JSON":
            msg = "[POST JSON] Calling API endpoint: {}".format(endpoint)
            self.log(msg)
            try:
                json_object = {}
                if post_json_tpl is not None and post_json_tpl.strip() != "":
                    json_object = json.loads(post_json_tpl)
                result = self.plugin.call_post_json(endpoint, json_object, headers)
            except Exception as e:
                msg = "Error: {}".format(e)
                self.error(e)
                self.log(msg)

        # GET
        elif command["type"] == "GET":
            msg = "[GET] Calling API endpoint: {}".format(endpoint)
            self.log(msg)
            result = self.plugin.call_get(endpoint, headers)

        if result is None:
            result = "No response from API."
            self.log(result)
        else:
            # encode bytes result to utf-8
            result = result.decode("utf-8")

        # prepare response
        request["url"] = endpoint
        response = {
            "request": request,
            "result": result,
        }
        return response



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.14 16:00:00                  #
# ================================================== #

import json
import os
import shutil
from packaging.version import Version

from pygpt_net.provider.core.preset.base import BaseProvider
from pygpt_net.item.preset import PresetItem
from .patch import Patch


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.patcher = Patch(window)
        self.id = "json_file"
        self.type = "preset"

    def install(self):
        """
        Install provider data
        """
        # install presets
        presets_dir = self.window.core.config.get_user_dir('presets')
        src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets')
        if not os.path.exists(presets_dir):
            shutil.copytree(src, presets_dir)
        else:
            # copy missing presets
            for file in os.listdir(src):
                src_file = os.path.join(src, file)
                dst_file = os.path.join(presets_dir, file)
                if not os.path.exists(dst_file):
                    shutil.copyfile(src_file, dst_file)

    def load(self) -> dict | None:
        """
        Load presets from JSON files

        :return: dict or None
        """
        items = {}
        path = self.window.core.config.get_user_dir('presets')
        if not os.path.exists(path):
            print("FATAL ERROR: {} not found!".format(path))
            return None
        try:
            for filename in os.listdir(path):
                if filename.endswith(".json"):
                    path = os.path.join(self.window.core.config.get_user_dir('presets'), filename)
                    with open(path, 'r', encoding="utf-8") as f:
                        preset = PresetItem()
                        self.deserialize(json.load(f), preset)
                        items[filename[:-5]] = preset
        except Exception as e:
            self.window.core.debug.log(e)

        return items

    def load_base(self) -> dict | None:
        """
        Load base presets from JSON files

        :return: dict or None
        """
        items = {}
        path = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets')
        if not os.path.exists(path):
            print("FATAL ERROR: {} not found!".format(path))
            return None
        try:
            for filename in os.listdir(path):
                if filename.endswith(".json"):
                    path = os.path.join(
                        os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets'),
                        filename
                    )
                    with open(path, 'r', encoding="utf-8") as f:
                        preset = PresetItem()
                        self.deserialize(json.load(f), preset)
                        items[filename[:-5]] = preset
        except Exception as e:
            self.window.core.debug.log(e)

        return items

    def save(self, id: str, item: PresetItem):
        """
        Save preset to JSON file

        :param id: preset id
        :param item: PresetItem
        """
        path = os.path.join(self.window.core.config.get_user_dir('presets'), id + '.json')
        data = self.serialize(item)
        data['__meta__'] = self.window.core.config.append_meta()
        dump = json.dumps(data, indent=4)
        try:
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)
        except Exception as e:
            self.window.core.debug.log(e)

    def save_all(self, items: dict):
        """
        Save all presets to JSON files

        :param items: items dict
        """
        for id in items:
            path = os.path.join(self.window.core.config.get_user_dir('presets'), id + '.json')

            # serialize
            data = self.serialize(items[id])
            data['__meta__'] = self.window.core.config.append_meta()
            dump = json.dumps(data, indent=4)
            try:
                with open(path, 'w', encoding="utf-8") as f:
                    f.write(dump)
            except Exception as e:
                self.window.core.debug.log(e)

    def remove(self, id: str):
        """
        Remove preset

        :param id: preset id
        """
        path = os.path.join(self.window.core.config.get_user_dir('presets'), id + '.json')
        if os.path.exists(path):
            try:
                os.remove(path)
            except Exception as e:
                self.window.core.debug.log(e)

    def truncate(self):
        pass

    def patch(self, version: Version) -> bool:
        """
        Migrate presets to current app version

        :param version: current app version
        :return: True if migrated
        """
        return self.patcher.execute(version)

    @staticmethod
    def serialize(item: PresetItem) -> dict:
        """
        Serialize item to dict

        :param item: item to serialize
        :return: serialized item
        """
        return {
            'uuid': item.uuid,
            'name': item.name,
            'ai_name': item.ai_name,
            'user_name': item.user_name,
            'prompt': item.prompt,
            'chat': item.chat,
            'completion': item.completion,
            'img': item.img,
            'vision': item.vision,
            'langchain': item.langchain,
            'assistant': item.assistant,
            "llama_index": item.llama_index,
            "agent": item.agent,
            "expert": item.expert,
            'temperature': item.temperature,
            'filename': item.filename,
            'model': item.model,
            'tools': item.tools,
            'experts': item.experts,
            'enabled': item.enabled,
        }

    @staticmethod
    def deserialize(data: dict, item: PresetItem):
        """
        Deserialize item from dict

        :param data: serialized item
        :param item: item to deserialize
        """
        if 'uuid' in data:
            item.uuid = data['uuid']
        if 'name' in data:
            item.name = data['name']
        if 'ai_name' in data:
            item.ai_name = data['ai_name']
        if 'user_name' in data:
            item.user_name = data['user_name']
        if 'prompt' in data:
            item.prompt = data['prompt']
        if 'chat' in data:
            item.chat = data['chat']
        if 'completion' in data:
            item.completion = data['completion']
        if 'img' in data:
            item.img = data['img']
        if 'vision' in data:
            item.vision = data['vision']
        if 'langchain' in data:
            item.langchain = data['langchain']
        if 'assistant' in data:
            item.assistant = data['assistant']
        if 'llama_index' in data:
            item.llama_index = data['llama_index']
        if 'agent' in data:
            item.agent = data['agent']
        if 'expert' in data:
            item.expert = data['expert']
        if 'temperature' in data:
            item.temperature = data['temperature']
        if 'filename' in data:
            item.filename = data['filename']
        if 'model' in data:
            item.model = data['model']
        if 'tools' in data:
            item.tools = data['tools']
        if 'experts' in data:
            item.experts = data['experts']
        if 'enabled' in data:
            item.enabled = data['enabled']

        # get version
        if '__meta__' in data and 'version' in data['__meta__']:
            item.version = data['__meta__']['version']

    def dump(self, item: PresetItem) -> str:
        """
        Dump to string

        :param item: item to dump
        :return: dumped item as string (json)
        """
        return json.dumps(self.serialize(item))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.26 23:00:00                  #
# ================================================== #

class Combo:
    def __init__(self, window=None):
        """
        Combobox field handler

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: dict
    ):
        """
        Apply value to combobox

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        :param idx: by idx, not the value
        """
        value = option["value"]
        if "idx" in option:  # by idx
            self.window.ui.config[parent_id][key].combo.setCurrentIndex(option["idx"])
        else: # by value
            index = self.window.ui.config[parent_id][key].combo.findData(value)
            if index != -1:
                self.window.ui.config[parent_id][key].combo.setCurrentIndex(index)

    def on_update(
            self,
            parent_id: str,
            key: str,
            option: dict,
            value: any,
            hooks: bool = True
    ):
        """
        Event: on change combobox value

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        :param value: Option value
        :param hooks: Run hooks
        """
        # on update hooks
        if hooks:
            hook_name = "update.{}.{}".format(parent_id, key)
            if self.window.ui.has_hook(hook_name):
                hook = self.window.ui.get_hook(hook_name)
                try:
                    hook(key, value, "combo")
                except Exception as e:
                    self.window.core.debug.log(e)

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: dict,
            idx: bool = False
    ) -> str or int:
        """
        Get checkbox value

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option data dict
        :param idx: return selected idx, not the value
        :return: Option text value or selected idx
        """
        if idx:
            return self.window.ui.config[parent_id][key].combo.currentIndex()
        else:
            return self.window.ui.config[parent_id][key].combo.currentData()

    def update_list(self, parent_id: str, key: str, items: dict):
        """
        Update combobox items

        :param parent_id: Options parent ID
        :param key: Option key
        :param items: Items dict
        """
        self.window.ui.config[parent_id][key].combo.clear()
        for item in items:
            self.window.ui.config[parent_id][key].combo.addItem(items[item], item)
        self.window.ui.config[parent_id][key].combo.setCurrentIndex(-1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.31 18:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class IndexList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Index select menu

        :param window: main window
        :param id: input id
        """
        super(IndexList, self).__init__(window)
        self.window = window
        self.id = id

    def click(self, val):
        self.window.controller.idx.select(val.row())

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: context menu event
        """
        actions = {}
        actions['edit'] = QAction(QIcon(":/icons/edit.svg"), trans('action.edit'), self)
        actions['edit'].triggered.connect(
            lambda: self.action_edit(event)
        )

        txt = trans('idx.index_now') + ': ' + trans('settings.llama.extra.btn.idx_db_all')
        actions['idx_db_all'] = QAction(QIcon(":/icons/search.db"), txt, self)
        actions['idx_db_all'].triggered.connect(
            lambda: self.action_idx_db_all(event)
        )

        txt = trans('idx.index_now') + ': ' + trans('settings.llama.extra.btn.idx_db_update')
        actions['idx_db_update'] = QAction(QIcon(":/icons/search.db"), txt, self)
        actions['idx_db_update'].triggered.connect(
            lambda: self.action_idx_db_update(event)
        )

        txt = trans('idx.index_now') + ': ' + trans('settings.llama.extra.btn.idx_files_all')
        actions['idx_files_all'] = QAction(QIcon(":/icons/search.db"), txt, self)
        actions['idx_files_all'].triggered.connect(
            lambda: self.action_idx_files_all(event)
        )

        actions['clear'] = QAction(QIcon(":/icons/close.svg"), trans('idx.btn.clear'), self)
        actions['clear'].triggered.connect(
            lambda: self.action_clear(event)
        )

        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('idx.btn.truncate'), self)
        actions['delete'].triggered.connect(
            lambda: self.action_truncate(event)
        )

        menu = QMenu(self)
        menu.addAction(actions['edit'])
        menu.addAction(actions['idx_db_all'])
        menu.addAction(actions['idx_db_update'])
        menu.addAction(actions['idx_files_all'])
        menu.addAction(actions['clear'])
        menu.addAction(actions['delete'])

        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.idx.select(item.row())
            menu.exec_(event.globalPos())

    def action_idx_db_all(self, event):
        """
        Idx action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.idx.idx_db_all_by_idx(idx)

    def action_idx_db_update(self, event):
        """
        Idx action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.idx.idx_db_update_by_idx(idx)

    def action_idx_files_all(self, event):
        """
        Idx action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.idx.idx_files_all_by_idx(idx)

    def action_edit(self, event):
        """
        Edit action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.open_section('llama-index')

    def action_clear(self, event):
        """
        Clear idx action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.idx.indexer.clear_by_idx(idx)

    def action_truncate(self, event):
        """
        Truncate idx action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.idx.indexer.truncate_by_idx(idx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.preset import PresetItem


class Experts:
    def __init__(self, window=None):
        """
        Experts core

        :param window: Window instance
        """
        self.window = window
        self.allowed_modes = ["chat", "completion", "vision", "langchain", "llama_index"]
        self.allowed_cmds = ["expert_call"]

    def get_mode(self) -> str:
        """
        Get sub-mode to use internally

        :return: sub-mode
        """
        mode = "chat"
        current = self.window.core.config.get("experts.mode")
        if current is not None and current in self.allowed_modes:
            mode = current
        return mode

    def stopped(self) -> bool:
        """
        Check if experts are stopped

        :return: True if stopped
        """
        return self.window.controller.agent.experts.stopped()

    def agent_enabled(self) -> bool:
        """
        Check if agent is enabled

        :return: True if enabled
        """
        return self.window.controller.agent.enabled()

    def get_expert(self, id: str) -> PresetItem:
        """
        Get expert by id

        :param id: expert id
        :return: expert item (preset)
        """
        return self.window.core.presets.get_by_id("expert", id)

    def get_experts(self) -> dict:
        """
        Get experts names with keys

        :return: experts dict
        """
        experts = {}
        presets = self.window.core.presets.get_by_mode("expert")

        # mode: agent
        if self.agent_enabled():
            agents = self.window.core.presets.get_by_mode("agent")
            agent = self.window.core.config.get('preset')
            if agent is not None:
                if agent in agents:
                    for uuid in agents[agent].experts:
                        expert = self.window.core.presets.get_by_uuid(uuid)
                        if expert is not None:
                            id = expert.filename
                            experts[id] = expert.name
        # mode: expert
        else:
            for k in presets:
                if k.startswith("current."):  # skip current
                    continue
                if not presets[k].enabled:  # skip disabled experts
                    continue
                experts[k] = presets[k].name
        return experts

    def get_prompt(self) -> str:
        """
        Get prompt for handle experts

        :return: prompt with list of experts appended
        """
        prompt = self.window.core.config.get('prompt.expert')
        experts = self.get_experts()
        experts_list = []
        for k in experts:
            if k.startswith("current."): # skip current presets
                continue
            experts_list.append(" - " + str(k) + ": " + str(experts[k]))
        prompt = prompt.replace("{presets}", "\n".join(experts_list))
        return prompt

    def extract_calls(self, ctx: CtxItem) -> dict:
        """
        Extract expert calls from context output

        :param ctx: context item
        :return: dict with calls
        """
        ids = self.get_experts().keys()
        if not ids:  # abort if no experts
            return {}
        cmds = self.window.core.command.extract_cmds(ctx.output)
        if len(cmds) > 0:
            ctx.cmds = cmds  # append commands to ctx
        else:  # abort if no cmds
            return {}
        commands = self.window.core.command.from_commands(cmds)  # pack to execution list
        is_cmd = False
        my_commands = []
        calls = {}
        for item in commands:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True
        if not is_cmd:  # abort if no expert calls
            return {}
        for item in my_commands:
            try:
                if item["cmd"] == "expert_call":
                    if "params" not in item:
                        continue
                    if "id" not in item["params"] or "query" not in item["params"]:
                        continue
                    id = item["params"]["id"]
                    if id not in ids:
                        continue
                    query = item["params"]["query"]
                    calls[id] = query
            except Exception as e:
                self.window.core.debug.error(e)
                return {}
        return calls

    def reply(self, ctx: CtxItem):
        """
        Re-send response from commands to master expert

        :param ctx: context item
        """
        if self.stopped():
            return

        internal = False
        if self.agent_enabled():  # hide in agent mode
            internal = True
        if ctx.output.strip() != "":
            response = ctx.output
        else:
            response = ctx.input
        self.window.controller.chat.input.send(
            "Result from expert:\n\n" + str(response),
            force=True,
            reply=True,
            internal=internal,
            prev_ctx=ctx,
        )

    def call(
            self,
            master_ctx: CtxItem,
            expert_id: str,
            query: str
    ):
        """
        Call the expert

        :param master_ctx: master context
        :param expert_id: expert id (preset ID)
        :param query: input text (master prompt)
        """
        if self.stopped():
            return

        # get or create children meta
        slave = self.window.core.ctx.get_or_create_slave_meta(master_ctx, expert_id)
        expert = self.get_expert(expert_id)
        reply = True
        hidden = False
        internal = False

        if self.agent_enabled():  # hide in agent mode
            internal = False
            hidden = True

        mode = self.get_mode()
        base_mode = mode
        model = expert.model
        user_name = expert.name
        ai_name = expert.name
        sys_prompt = expert.prompt
        model_data = self.window.core.models.get(model)

        files = []
        file_ids = []
        functions = []
        tools_outputs = []

        # from current config
        max_tokens = self.window.core.config.get('max_output_tokens')
        stream_mode = self.window.core.config.get('stream')

        # create slave item
        ctx = CtxItem()
        ctx.meta_id = slave.id
        ctx.internal = internal
        ctx.hidden = hidden
        ctx.current = True  # mark as current context item
        ctx.mode = mode  # store current selected mode (not inline changed)
        ctx.model = model  # store model list key, not real model id
        ctx.set_input(query, user_name)
        ctx.set_output(None, str(ai_name))
        ctx.sub_call = True  # mark as sub-call

        # render: begin
        self.window.controller.chat.render.begin(stream=stream_mode)
        self.window.core.ctx.provider.append_item(slave, ctx)  # to slave meta

        # build sys prompt
        sys_prompt_raw = sys_prompt  # store raw prompt
        event = Event(Event.PRE_PROMPT, {
            'mode': mode,
            'value': sys_prompt,
        })
        self.window.core.dispatcher.dispatch(event)
        sys_prompt = event.data['value']
        sys_prompt = self.window.core.prompt.prepare_sys_prompt(
            mode,
            sys_prompt,
            ctx,
            reply,
            internal,
        )

        # call bridge
        history = self.window.core.ctx.all(
            meta_id=slave.id
        )  # get history for slave ctx
        bridge_context = BridgeContext(
            ctx=ctx,
            history=history,
            mode=mode,
            parent_mode=base_mode,
            model=model_data,
            system_prompt=sys_prompt,
            system_prompt_raw=sys_prompt_raw,
            prompt=query,
            stream=stream_mode,
            attachments=files,
            file_ids=file_ids,
            assistant_id=self.window.core.config.get('assistant'),
            idx=self.window.controller.idx.current_idx,
            idx_raw=self.window.core.config.get('llama.idx.raw'),
            external_functions=functions,
            tools_outputs=tools_outputs,
            max_tokens=max_tokens,
        )
        self.window.controller.chat.common.lock_input()  # lock input
        result = self.window.core.bridge.call(
            context=bridge_context,
        )
        if not result:  # abort if bridge call failed
            return

        # handle output
        ctx.current = False  # reset current state
        self.window.core.ctx.update_item(ctx)
        ctx.from_previous()  # append previous result if exists
        self.window.controller.chat.output.handle(
            ctx,
            mode,
            stream_mode,
        )
        ctx.clear_reply()  # reset results
        self.window.controller.chat.output.handle_cmd(ctx)  # handle cmds
        ctx.from_previous()  # append previous result again before save
        self.window.core.ctx.update_item(ctx)  # update ctx in DB

        # if commands reply after bridge call, then stop (already handled in dispatcher)
        if ctx.reply:
            return

        # send slave expert response to master expert
        self.window.controller.chat.input.send(
            "@"+expert_id+" says:\n\n" + str(ctx.output),
            force=True,
            reply=False,
            internal=False,
            prev_ctx=ctx,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

import sys
import argparse
from logging import ERROR, WARNING, INFO, DEBUG

from PySide6.QtCore import QCoreApplication, Qt
from PySide6.QtGui import QScreen
from PySide6.QtWidgets import QApplication

from pygpt_net.core.access.events import AppEvent
from pygpt_net.core.access.shortcuts import GlobalShortcutFilter
from pygpt_net.core.debug import Debug
from pygpt_net.core.platforms import Platforms
from pygpt_net.tools import BaseTool
from pygpt_net.ui.main import MainWindow
from pygpt_net.plugin.base import BasePlugin
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.provider.loaders.base import BaseLoader
from pygpt_net.provider.vector_stores.base import BaseStore
from pygpt_net.provider.audio_input.base import BaseProvider as BaseAudioInput
from pygpt_net.provider.audio_output.base import BaseProvider as BaseAudioOutput
from pygpt_net.provider.web.base import BaseProvider as BaseWeb


class Launcher:
    def __init__(self):
        """Launcher"""
        self.app = None
        self.window = None
        self.debug = False
        self.force_legacy = False
        self.force_disable_gpu = False
        self.shortcut_filter = None

    def setup(self) -> dict:
        """
        Setup launcher

        :return: dict with launcher arguments
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-d",
            "--debug",
            required=False,
            help="debug mode (0=disabled, 1=info, 2=debug)",
        )
        parser.add_argument(
            "-l",
            "--legacy",
            required=False,
            help="force enable legacy mode (0=disabled, 1=enable)",
        )
        parser.add_argument(
            "-n",
            "--disable-gpu",
            required=False,
            help="force disable OpenGL (1=disabled, 0=enabled)",
        )
        args = vars(parser.parse_args())

        # set log level [ERROR|WARNING|INFO|DEBUG]
        if "debug" in args and args["debug"] == "1":
            print("** Debug mode enabled (1=INFO)")
            Debug.init(INFO)
            self.debug = True
        elif "debug" in args and args["debug"] == "2":
            print("** Debug mode enabled (2=DEBUG)")
            Debug.init(DEBUG)
            self.debug = True
        else:
            Debug.init(ERROR)  # default log level

        # force legacy mode
        if "legacy" in args and args["legacy"] == "1":
            print("** Force legacy mode enabled")
            self.force_legacy = True

        # force disable GPU
        if "disable-gpu" in args and args["disable-gpu"] == "1":
            print("** Force disable GPU enabled")
            self.force_disable_gpu = True

        return args

    def init(self):
        """Initialize app"""
        args = self.setup()
        Platforms.prepare()  # setup platform specific options
        QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts)
        self.app = QApplication(sys.argv)
        self.window = MainWindow(self.app, args=args)
        self.shortcut_filter = GlobalShortcutFilter(self.window)

    def add_plugin(self, plugin: BasePlugin):
        """
        Register plugin

        :param plugin: plugin instance
        """
        if not isinstance(plugin, BasePlugin):
            raise TypeError(
                "Plugin must be instance of: "
                "pygpt_net.plugin.base.BasePlugin"
            )
        self.window.add_plugin(plugin)
        if self.debug:
            print("Loaded plugin: {} ({})".format(plugin.id, plugin.__class__.__name__))

    def add_llm(self, llm: BaseLLM):
        """
        Register LLM provider

        :param llm: LLM provider instance
        """
        if not isinstance(llm, BaseLLM):
            raise TypeError(
                "LLM provider must be instance of: "
                "pygpt_net.provider.llms.base.BaseLLM"
            )
        self.window.add_llm(llm)
        if self.debug:
            print("Loaded LLM: {} ({})".format(llm.id, llm.__class__.__name__))

    def add_vector_store(self, store: BaseStore):
        """
        Register vector store provider

        :param store: Vector store provider instance
        """
        if not isinstance(store, BaseStore):
            raise TypeError(
                "Vector store provider must be instance of: "
                "pygpt_net.provider.vector_stores.base.BaseStore"
            )
        self.window.add_vector_store(store)
        if self.debug:
            print("Loaded vector store: {} ({})".format(store.id, store.__class__.__name__))

    def add_loader(self, loader: BaseLoader):
        """
        Register data loader

        :param loader: Data loader instance
        """
        if not isinstance(loader, BaseLoader):
            raise TypeError(
                "Data loader must be instance of: "
                "pygpt_net.provider.loaders.base.BaseLoader"
            )
        self.window.add_loader(loader)
        if self.debug:
            print("Loaded data loader: {} ({})".format(loader.id, loader.__class__.__name__))

    def add_audio_input(self, audio: BaseAudioInput):
        """
        Register audio input provider

        :param audio: Audio input provider instance
        """
        if not isinstance(audio, BaseAudioInput):
            raise TypeError(
                "Audio input provider must be instance of: "
                "pygpt_net.provider.audio_input.base.BaseProvider"
            )
        self.window.add_audio_input(audio)
        if self.debug:
            print("Loaded audio input: {} ({})".format(audio.id, audio.__class__.__name__))

    def add_audio_output(self, audio: BaseAudioOutput):
        """
        Register audio output provider

        :param audio: Audio output provider instance
        """
        if not isinstance(audio, BaseAudioOutput):
            raise TypeError(
                "Audio output provider must be instance of: "
                "pygpt_net.provider.audio_output.base.BaseProvider"
            )
        self.window.add_audio_output(audio)
        if self.debug:
            print("Loaded audio output: {} ({})".format(audio.id, audio.__class__.__name__))

    def add_web(self, provider: BaseWeb):
        """
        Register web provider

        :param provider: Web provider instance
        """
        if not isinstance(provider, BaseWeb):
            raise TypeError(
                "Web provider must be instance of: "
                "pygpt_net.provider.web.base.BaseProvider"
            )
        self.window.add_web(provider)
        if self.debug:
            print("Loaded web provider: {} ({})".format(provider.id, provider.__class__.__name__))

    def add_tool(self, tool: BaseTool):
        """
        Register tool

        :param tool: tool instance
        """
        if not isinstance(tool, BaseTool):
            raise TypeError(
                "Tool must be instance of: "
                "pygpt_net.tools.base.BaseTool"
            )
        self.window.add_tool(tool)
        if self.debug:
            print("Loaded tool: {} ({})".format(tool.id, tool.__class__.__name__))

    def run(self):
        """Run app"""
        self.window.setup()
        geometry = self.window.screen().availableGeometry()
        pos = QScreen.availableGeometry(QApplication.primaryScreen()).topLeft()
        margin = 100
        self.window.resize(geometry.width() - margin, geometry.height() - margin)
        self.window.show()
        self.window.move(pos)
        self.window.post_setup()
        self.app.setWindowIcon(self.window.ui.get_app_icon())
        self.window.ui.tray.setup(self.app)
        self.window.controller.after_setup()
        self.window.core.dispatcher.dispatch(AppEvent(AppEvent.APP_STARTED))  # app event
        self.window.installEventFilter(self.shortcut_filter)
        sys.exit(self.app.exec())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 16:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QPushButton, QMenu

from pygpt_net.utils import trans


class ContextMenuButton(QPushButton):
    def __init__(self, title, parent=None, action=None):
        super().__init__(title, parent)
        self.action = action

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton or event.button() == Qt.RightButton:
            self.action(self, event.pos())
        else:
            super().mousePressEvent(event)


class NewCtxButton(QPushButton):
    def __init__(self, title: str = None, window=None):
        super().__init__(title)
        self.window = window
        self.clicked.connect(
            lambda: self.window.controller.ctx.new()
        )

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            self.new_context_menu(self, event.pos())
        else:
            super().mousePressEvent(event)

    def new_context_menu(self, parent, pos):
        """
        Index all btn context menu

        :param parent: parent widget
        :param pos: mouse  position
        """
        group_id = self.window.controller.ctx.group_id
        menu = QMenu(self)
        actions = {}
        actions['new'] = QAction(QIcon(":/icons/add.svg"), trans('action.ctx.new'), self)
        actions['new'].triggered.connect(
            lambda: self.window.controller.ctx.new_ungrouped()
        )
        if group_id is not None and group_id > 0:
            group_name = self.window.controller.ctx.get_group_name(group_id)
            actions['new_in_group'] = QAction(QIcon(":/icons/add.svg"), trans('action.ctx.new.in_group').format(group=group_name), self)
            actions['new_in_group'].triggered.connect(
                lambda checked=False, id=group_id: self.window.controller.ctx.new(force=False, group_id=id)
            )
        actions['new_group'] = QAction(QIcon(":/icons/folder_filled.svg"), trans('menu.file.group.new'), self)
        actions['new_group'].triggered.connect(
            lambda: self.window.controller.ctx.new_group()
        )

        if group_id is not None and group_id > 0:
            menu.addAction(actions['new_in_group'])
        menu.addAction(actions['new'])
        menu.addAction(actions['new_group'])
        menu.exec_(parent.mapToGlobal(pos))


class SyncButton(QPushButton):
    def __init__(self, title: str = None, window=None):
        super().__init__(title)
        self.window = window
        self.clicked.connect(
            lambda: self.window.controller.assistant.batch.import_files_current()
        )

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            self.new_context_menu(self, event.pos())
        else:
            super().mousePressEvent(event)

    def new_context_menu(self, parent, pos):
        """
        Index all btn context menu

        :param parent: parent widget
        :param pos: mouse  position
        """
        menu = QMenu(self)
        actions = {}
        actions['current'] = QAction(QIcon(":/icons/download.svg"), trans('attachments_uploaded.btn.sync.current'), self)
        actions['current'].triggered.connect(
            lambda: self.window.controller.assistant.batch.import_files_current()
        )
        actions['all'] = QAction(QIcon(":/icons/download.svg"), trans('attachments_uploaded.btn.sync.all'), self)
        actions['all'].triggered.connect(
            lambda: self.window.controller.assistant.batch.import_files()
        )
        menu.addAction(actions['current'])
        menu.addAction(actions['all'])
        menu.exec_(parent.mapToGlobal(pos))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.25 01:00:00                  #
# ================================================== #

from pygpt_net.item.ctx import CtxItem


class BaseRenderer:
    def __init__(self, window=None):
        """
        Base renderer

        :param window: Window instance
        """
        self.window = window

    def is_stream(self) -> bool:
        """
        Check if it is a stream

        :return: True if stream is enabled
        """
        return self.window.core.config.get("stream")

    def begin(self, stream: bool = False):
        """Render begin"""
        pass

    def end(self, stream: bool = False):
        """Render end"""
        pass

    def end_extra(self, stream: bool = False):
        """Render end extra"""
        pass

    def stream_begin(self):
        """Render stream begin"""
        pass  # do nothing

    def stream_end(self):
        """Render stream end"""
        pass  # do nothing    

    def clear_output(self):
        """Clear output"""
        pass

    def clear_input(self):
        """Clear input"""
        pass

    def reset(self):
        """Reset"""
        pass

    def reload(self):
        """Reload output, called externally only on theme change to redraw content"""
        pass

    def append_context(self, items: list, clear: bool = True):
        """
        Append all context to output

        :param items: Context items
        :param clear: True if clear all output before append
        """
        pass

    def append_input(self, item: CtxItem):
        """
        Append text input to output

        :param item: context item
        """
        pass

    def append_output(self, item: CtxItem):
        """
        Append text output to output

        :param item: context item
        """
        pass

    def append_extra(self, item: CtxItem, footer: bool = False):
        """
        Append extra data (images, files, etc.) to output

        :param item: context item
        :param footer: True if it is a footer
        """
        pass

    def append_chunk(self, item: CtxItem, text_chunk: str, begin: bool = False):
        """
        Append output chunk to output

        :param item: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        pass

    def remove_item(self, id: int):
        """
        Remove item from output

        :param id: context item ID
        """
        pass

    def remove_items_from(self, id: int):
        """
        Remove item from output

        :param id: context item ID
        """
        pass

    def on_edit_submit(self, id: int):
        """
        On edit submit

        :param id: context item ID
        """
        self.window.controller.ctx.refresh(restore_model=False)  # allow model change

    def on_remove_submit(self, id: int):
        """
        On remove submit

        :param id: context item ID
        """
        pass

    def on_reply_submit(self, id: int):
        """
        On regenerate submit

        :param id: context item ID
        """
        self.window.controller.ctx.refresh(restore_model=False)  # allow model change

    def clear_all(self):
        """Clear all"""
        pass

    def on_page_loaded(self):
        """On page loaded callback"""
        pass

    def on_enable_edit(self, live: bool = True):
        """
        On enable edit icons

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_disable_edit(self, live: bool = True):
        """
        On disable edit icons

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_enable_timestamp(self, live: bool = True):
        """
        On enable timestamp

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_disable_timestamp(self, live: bool = True):
        """
        On disable timestamp

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_theme_change(self):
        """On theme change"""
        pass

    def get_scroll_position(self) -> int:
        """
        Get scroll position

        :return: scroll position
        """
        return 0

    def set_scroll_position(self, position: int):
        """
        Set scroll position

        :param position: scroll position
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from pygpt_net.core.dispatcher import BaseEvent
from pygpt_net.core.access.events import ControlEvent, AppEvent

from .control import Control
from .voice import Voice


class Access:
    def __init__(self, window=None):
        """
        Accessibility controller

        :param window: Window instance
        """
        self.window = window
        self.control = Control(window)
        self.voice = Voice(window)

    def setup(self):
        """Setup accessibility"""
        self.voice.setup()

    def reload(self):
        """Reload accessibility"""
        self.voice.setup()

    def update(self):
        """Update accessibility"""
        self.voice.update()

    def handle(self, event: BaseEvent):
        """
        Handle accessibility event

        :param event: event object
        """
        if isinstance(event, ControlEvent):
            self.control.handle(event)
            event.stop = True
        elif isinstance(event, AppEvent):
            self.handle_app(event)
            event.stop = True

    def handle_app(self, event: AppEvent):
        """
        Handle accessibility event (app)

        :param event: event object
        """
        self.window.core.debug.info("EVENT APP: " + event.name)
        if event.name == AppEvent.VOICE_CONTROL_TOGGLE:
            self.voice.toggle_recording()
        self.voice.play(event)  # handle audio events

    def on_escape(self):
        """Handle escape key"""
        if self.voice.is_recording:
            self.voice.stop_recording(timeout=True)
        if self.window.core.plugins.get("audio_input").handler_simple.is_recording:
            self.window.core.plugins.get("audio_input").handler_simple.stop_recording(timeout=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.16 15:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot

from pygpt_net.plugin.base import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = BaseSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        msg = None
        for item in self.cmds:
            response = None
            try:
                if item["cmd"] == "get_context":
                    response = self.cmd_get_context(item)

                if response:
                    responses.append(response)

            except Exception as e:
                msg = "Error: {}".format(e)
                responses.append({
                    "request": {
                        "cmd": item["cmd"],
                    },
                    "result": "Error {}".format(e),
                })
                self.error(e)
                self.log(msg)

        # send response
        if len(responses) > 0:
            for response in responses:
                self.reply(response)

        # update status
        if msg is not None:
            self.status(msg)

    def cmd_get_context(self, item: dict) -> dict:
        """
        Get context for given query

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        question = item["params"]["query"]
        self.status("Please wait... querying: {}...".format(question))
        answer, doc_ids, metas = self.plugin.query(question)  # send question to Llama-index
        response = {
            "request": request,
            "result": answer,
            "doc_ids": doc_ids,
            "metas": metas,
            "context": "ADDITIONAL CONTEXT:\n--------------------------------\n" + answer,
        }
        if doc_ids:
            self.ctx.doc_ids = doc_ids  # store doc_ids in context
        return response

    def prepare_request(self, item) -> dict:
        """
        Prepare request item for result

        :param item: item with parameters
        :return: request item
        """
        return {"cmd": item["cmd"]}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 03:00:00                  #
# ================================================== #

from llama_index.core.llms import ChatMessage, MessageRole


class Context:
    def __init__(self, window=None):
        """
        Context core

        :param window: Window instance
        """
        self.window = window

    def get_messages(
            self,
            input_prompt: str,
            system_prompt: str,
            history: list = None
    ):
        """
        Get messages from db

        :param input_prompt: input prompt
        :param system_prompt: system prompt
        :param history: history
        :return: Messages
        """
        messages = []

        # tokens config
        model = self.window.core.config.get('model')
        model_id = self.window.core.models.get_id(model)
        mode = self.window.core.config.get('mode')

        used_tokens = self.window.core.tokens.from_user(input_prompt, system_prompt)  # threshold and extra included
        max_tokens = self.window.core.config.get('max_total_tokens')
        model_ctx = self.window.core.models.get_num_ctx(model_id)

        # fit to max model tokens
        if max_tokens > model_ctx:
            max_tokens = model_ctx

        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model_id,
                mode,
                used_tokens,
                max_tokens,
            )
            for item in items:
                # input
                if item.input is not None and item.input != "":
                    messages.append(ChatMessage(
                        role=MessageRole.USER,
                        content=item.input
                    ))
                # output
                if item.output is not None and item.output != "":
                    messages.append(ChatMessage(
                        role=MessageRole.ASSISTANT,
                        content=item.output
                    ))

        return messages



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.01 02:00:00                  #
# ================================================== #

import json
import os
from typing import Any

from llama_index.readers.google.keep.base import GoogleKeepReader as Reader

class GoogleKeepReader(Reader):
    """
    Wrapper for:

    llama_index.readers.google.keep.base.GoogleKeepReader

    Extending base reader with credentials configuration.
    """

    def __init__(
        self,
        credentials_path: str = "keep_credentials.json",
    ) -> None:
        """Initialize with parameters."""
        self.credentials_path = credentials_path

    def _get_keep(self) -> Any:
        import gkeepapi

        """Get a Google Keep object with login."""
        # Read username and password from keep_credentials.json
        if os.path.exists(self.credentials_path):
            with open(self.credentials_path) as f:
                credentials = json.load(f)
        else:
            raise RuntimeError("Failed to load " + str(self.credentials_path) + " file.")

        keep = gkeepapi.Keep()
        success = keep.login(credentials["username"], credentials["password"])
        if not success:
            raise RuntimeError("Failed to login to Google Keep.")

        return keep


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.25 19:00:00                  #
# ================================================== #

from PySide6.QtCore import QObject, Signal, Slot
from pygpt_net.core.worker import Worker
from pygpt_net.item.ctx import CtxItem


class WorkerSignals(QObject):
    updated = Signal(int, object, str)


class Summarizer:
    def __init__(self, window=None):
        """
        Summarize  controller

        :param window: Window instance
        """
        self.window = window

    def summarize(self, id: int, ctx: CtxItem):
        """
        Summarize context

        :param id: CtxMeta ID
        :param ctx: CtxItem
        """
        self.start_worker(id, ctx)

    def summarizer(self, id: int, ctx: CtxItem, window, updated_signal: Signal):
        """
        Summarize worker callback

        :param id: CtxMeta ID
        :param ctx: CtxItem
        :param window: Window instance
        :param updated_signal: WorkerSignals: updated signal
        """
        title = window.core.gpt.summarizer.summary_ctx(ctx)
        if title:
            updated_signal.emit(id, ctx, title)

    def start_worker(self, id: int, ctx: CtxItem):
        """
        Handle worker thread

        :param id: CtxMeta ID
        :param ctx: CtxItem
        """
        worker = Worker(self.summarizer)
        worker.signals = WorkerSignals()
        worker.signals.updated.connect(self.handle_update)
        worker.kwargs['id'] = id
        worker.kwargs['ctx'] = ctx
        worker.kwargs['window'] = self.window
        worker.kwargs['updated_signal'] = worker.signals.updated
        self.window.threadpool.start(worker)

    @Slot(int, object, str)
    def handle_update(self, id: int, ctx: CtxItem, title: str):
        """
        Handle update signal (make update)

        :param id: CtxMeta ID
        :param ctx: CtxItem
        :param title: CtxMeta title
        """
        refresh = True
        # prevent UI list selection loose after later update
        if ctx.internal or len(ctx.cmds) > 0:
            refresh = False
        self.window.controller.ctx.update_name(
            id,
            title,
            refresh=refresh,
        )




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.17 01:00:00                  #
# ================================================== #

from llama_index.core.readers.base import BaseReader


class BaseLoader:
    def __init__(self, *args, **kwargs):
        self.window = None
        self.id = ""
        self.name = ""
        self.extensions = []
        self.type = ["file"]  # list of types: file, web
        self.instructions = []  # list of instructions for 'web_index' command for how to handle this type
        self.args = {}  # custom keyword arguments
        self.init_args = {}  # initial keyword arguments
        self.init_args_types = {}
        self.allow_compiled = True  # allow in compiled and Snap versions
        # This is required due to some readers may require Python environment to install additional packages

    def attach_window(self, window):
        """
        Attach window instance

        :param window: Window instance
        """
        self.window = window

    def set_args(self, args: dict):
        """
        Set loader initial keyword arguments

        :param args: keyword arguments dict
        """
        self.args = args

    def get_args(self):
        """
        Prepare keyword arguments for reader init method

        :return: keyword arguments dict
        """
        args = {}
        for key in self.init_args:
            args[key] = self.init_args[key]
            if key in self.args:
                args[key] = self.args[key]
        return args

    def prepare_args(self, **kwargs) -> dict:
        """
        Prepare arguments for reader load method

        :param kwargs: keyword arguments
        :return: args to pass to reader
        """
        return kwargs

    def get_external_id(self, args: dict = None) -> str:
        """
        Get unique web content identifier

        :param args: load_data args
        :return: unique content identifier
        """
        if "url" in args:
            return args.get("url")
        return ""

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

class Common:
    def __init__(self, window=None):
        """
        Agent common controller

        :param window: Window instance
        """
        self.window = window

    def enable_auto_stop(self):
        """Enable auto stop"""
        self.window.core.config.set('agent.auto_stop', True)
        self.window.core.config.save()

    def disable_auto_stop(self):
        """Disable auto stop"""
        self.window.core.config.set('agent.auto_stop', False)
        self.window.core.config.save()

    def toggle_auto_stop(self, state: bool):
        """
        Toggle auto stop

        :param state: state of checkbox
        """
        if not state:
            self.disable_auto_stop()
        else:
            self.enable_auto_stop()

    def show_status(self):
        """Show agent status"""
        self.window.ui.nodes['status.agent'].setVisible(True)

    def hide_status(self):
        """Hide agent status"""
        self.window.ui.nodes['status.agent'].setVisible(False)

    def toggle_status(self):
        """Toggle agent status"""
        mode = self.window.core.config.get('mode')
        if mode == 'agent' or self.window.controller.agent.is_inline():
            self.show_status()
        else:
            self.hide_status()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from .events import ControlEvent


class Actions:
    def __init__(self, window=None):
        """
        Actions core

        :param window: Window instance
        """
        self.window = window

    def get_access_choices(self) -> list:
        """
        Get app access events choices

        :return: choices
        """
        events_list = [
            ControlEvent.APP_STATUS,
            ControlEvent.APP_EXIT,
            ControlEvent.AUDIO_OUTPUT_ENABLE,
            ControlEvent.AUDIO_OUTPUT_DISABLE,
            ControlEvent.AUDIO_INPUT_ENABLE,
            ControlEvent.AUDIO_INPUT_DISABLE,
            ControlEvent.CAMERA_ENABLE,
            ControlEvent.CAMERA_DISABLE,
            ControlEvent.CAMERA_CAPTURE,
            ControlEvent.CTX_NEW,
            ControlEvent.CTX_PREV,
            ControlEvent.CTX_NEXT,
            ControlEvent.CTX_LAST,
            ControlEvent.CTX_INPUT_FOCUS,
            ControlEvent.CTX_INPUT_SEND,
            ControlEvent.CTX_INPUT_CLEAR,
            ControlEvent.CTX_STOP,
            ControlEvent.CTX_ATTACHMENTS_CLEAR,
            ControlEvent.MODE_CHAT,
            ControlEvent.MODE_LLAMA_INDEX,
            ControlEvent.TAB_CHAT,
            ControlEvent.TAB_CALENDAR,
            ControlEvent.TAB_DRAW,
            ControlEvent.TAB_FILES,
            ControlEvent.TAB_NOTEPAD,
            ControlEvent.TAB_NEXT,
            ControlEvent.TAB_PREV,
            ControlEvent.VOICE_MESSAGE_START,
            ControlEvent.VOICE_MESSAGE_STOP,
            ControlEvent.VOICE_MESSAGE_TOGGLE,
            ControlEvent.VOICE_COMMAND_START,
            ControlEvent.VOICE_COMMAND_STOP,
            ControlEvent.VOICE_COMMAND_TOGGLE,
        ]
        # build choices
        choices = []
        for event in events_list:
            choices.append({event: f"event.control.{event}"})
        return choices


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.26 23:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QTextEdit


class OptionTextarea(QTextEdit):
    def __init__(
            self,
            window=None,
            parent_id: str = None,
            id: str = None,
            option: dict = None
    ):
        """
        Settings textarea

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(OptionTextarea, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option
        self.value = False
        self.title = ""
        self.real_time = False
        self.update_ui = True
        self.setAcceptRichText(False)
        self.context_options = []

        # init from option data
        if self.option is not None:
            if "label" in self.option:
                self.title = self.option["label"]
            if "value" in self.option:
                self.value = self.option["value"]
            if "real_time" in self.option:
                self.real_time = self.option["real_time"]
            if "context_options" in self.option:
                self.context_options = self.option["context_options"]
            if "read_only" in self.option and self.option["read_only"]:
                self.setReadOnly(True)

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: Event
        """
        menu = self.createStandardContextMenu()
        if "prompt.template.paste" in self.context_options:
            self.window.core.prompt.template.to_menu_options(menu, "editor")
        menu.exec_(event.globalPos())

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        super(OptionTextarea, self).keyPressEvent(event)
        if not self.real_time:
            return

        # hook update
        self.window.controller.config.input.on_update(
            self.parent_id,
            self.id,
            self.option,
            self.toPlainText()
        )
        if self.update_ui:
            self.window.controller.ui.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 07:00:00                  #
# ================================================== #

from .note import Note


class Calendar:
    def __init__(self, window=None):
        """
        Calendar controller

        :param window: Window instance
        """
        self.window = window
        self.note = Note(window)
        self.selected_year = None
        self.selected_month = None
        self.selected_day = None

    def setup(self):
        """Setup calendar"""
        self.note.setup()
        self.load()
        self.update()  # update counters and load notes for current month
        self.set_current()  # set to current note at start

    def update(self, all: bool = True):
        """
        Update counters

        :param all: reload all notes
        """
        year = self.window.ui.calendar['select'].currentYear
        month = self.window.ui.calendar['select'].currentMonth
        self.on_page_changed(year, month, all=all)  # load notes for current month

    def update_ctx_counters(self):
        """Update context counters only"""
        year = self.window.ui.calendar['select'].currentYear
        month = self.window.ui.calendar['select'].currentMonth
        self.note.refresh_ctx(year, month)

    def set_current(self):
        """Set to current selected date"""
        year = self.window.ui.calendar['select'].currentYear
        month = self.window.ui.calendar['select'].currentMonth
        day = self.window.ui.calendar['select'].currentDay

        self.note.update_content(year, month, day)
        self.note.update_label(year, month, day)

        self.selected_year = year
        self.selected_month = month
        self.selected_day = day

    def load(self):
        """Load notes from current year and month from database"""
        year = self.window.ui.calendar['select'].currentYear
        month = self.window.ui.calendar['select'].currentMonth
        self.window.core.calendar.load_by_month(year, month)

    def on_page_changed(self, year: int, month: int, all: bool = True):
        """
        On calendar page changed

        :param year: year
        :param month: month
        :param all: reload all notes
        """
        if all:
            self.load()  # reload notes for current year and month
        self.note.refresh_ctx(year, month)
        self.note.refresh_num(year, month)

    def on_day_select(self, year: int, month: int, day: int):
        """
        On day select

        :param year: year
        :param month: month
        :param day: day
        """
        self.selected_year = year
        self.selected_month = month
        self.selected_day = day
        self.note.update_content(year, month, day)
        self.note.update_label(year, month, day)

    def on_ctx_select(self, year: int, month: int, day: int):
        """
        On ctx select

        :param year: year
        :param month: month
        :param day: day
        """
        search_string = '@date({:04d}-{:02d}-{:02d})'.format(year, month, day)
        self.window.controller.ctx.append_search_string(search_string)

    def save_all(self):
        """Save all calendar notes"""
        self.window.core.calendar.save_all()

    def is_active(self) -> bool:
        """
        Check if calendar tab is active

        :return: True if calendar tab is active
        """
        return self.window.ui.tabs['output'].currentIndex() == self.window.controller.ui.tab_idx['calendar']

    def reload(self):
        """Reload calendar"""
        self.setup()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.30 15:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtGui import QStandardItemModel, Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QPushButton, QWidget, QCheckBox, QSizePolicy

from pygpt_net.ui.widget.element.labels import HelpLabel, TitleLabel
from pygpt_net.ui.widget.lists.index import IndexList
from pygpt_net.ui.widget.lists.index_combo import IndexCombo
from pygpt_net.utils import trans


class Indexes:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'indexes'

    def setup(self) -> QWidget:
        """
        Setup indexes

        :return: QWidget
        """
        layout = self.setup_idx()

        self.window.ui.nodes['indexes.widget'] = QWidget()
        self.window.ui.nodes['indexes.widget'].setLayout(layout)
        self.window.ui.nodes['indexes.widget'].setMinimumHeight(150)

        return self.window.ui.nodes['indexes.widget']

    def setup_idx(self) -> QVBoxLayout:
        """
        Setup list of indexes

        :return: QVBoxLayout
        """
        # new
        self.window.ui.nodes['indexes.new'] = QPushButton(trans('idx.new'))
        self.window.ui.nodes['indexes.new'].clicked.connect(
            lambda: self.window.controller.settings.open_section('llama-index'))

        # label
        self.window.ui.nodes['indexes.label'] = TitleLabel(trans("toolbox.indexes.label"))

        # header
        header = QHBoxLayout()
        header.addWidget(self.window.ui.nodes['indexes.label'])
        header.addStretch(1)
        header.addWidget(self.window.ui.nodes['indexes.new'], alignment=Qt.AlignRight)
        header.setContentsMargins(0, 0, 0, 0)
        header_widget = QWidget()
        header_widget.setLayout(header)

        # list
        self.window.ui.nodes[self.id] = IndexList(self.window, self.id)
        self.window.ui.nodes[self.id].selection_locked = self.window.controller.idx.change_locked
        self.window.ui.nodes[self.id].setMinimumWidth(40)

        self.window.ui.nodes['tip.toolbox.indexes'] = HelpLabel(trans('tip.toolbox.indexes'), self.window)

        # rows
        layout = QVBoxLayout()
        layout.addWidget(header_widget)
        layout.addWidget(self.window.ui.nodes[self.id])
        #layout.addWidget(self.window.ui.nodes['tip.toolbox.indexes'])

        # model
        self.window.ui.models[self.id] = self.create_model(self.window)
        self.window.ui.nodes[self.id].setModel(self.window.ui.models[self.id])

        return layout

    def setup_options(self) -> QWidget:
        """
        Setup idx options

        :return: QWidget
        :rtype: QWidget
        """
        # idx query only
        self.window.ui.config['global']['llama.idx.raw'] = QCheckBox(trans("idx.query.raw"))
        self.window.ui.config['global']['llama.idx.raw'].stateChanged.connect(
            lambda: self.window.controller.idx.common.toggle_raw(
                self.window.ui.config['global']['llama.idx.raw'].isChecked()
            )
        )

        # label
        # label = QLabel(trans("toolbox.llama_index.label"))

        # add options
        cols = QHBoxLayout()
        cols.addWidget(self.window.ui.config['global']['llama.idx.raw'])

        # index combo
        option = {
            "name": "current_index",
            "label": "toolbox.llama_index.current_index",
            "keys": [],
            "value": None,
        }
        self.window.ui.nodes['indexes.select'] = IndexCombo(
            self.window,
            'global',
            'current_index',
            option
        )
        self.window.ui.nodes['indexes.select'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # tip
        # self.window.ui.nodes['tip.toolbox.indexes'] = HelpLabel(trans('tip.toolbox.indexes'), self.window)

        # new
        self.window.ui.nodes['indexes.new'] = QPushButton(trans('idx.new'))
        self.window.ui.nodes['indexes.new'].clicked.connect(
            lambda: self.window.controller.settings.open_section('llama-index'))

        # label
        self.window.ui.nodes['indexes.label'] = TitleLabel(trans("toolbox.indexes.label"))

        idx_layout = QHBoxLayout()
        idx_layout.addWidget(self.window.ui.nodes['indexes.label'])
        idx_layout.addWidget(self.window.ui.nodes['indexes.select'])
        idx_layout.addWidget(self.window.ui.nodes['indexes.new'], alignment=Qt.AlignRight)
        idx_layout.setContentsMargins(0, 0, 0, 10)
        idx_widget = QWidget()
        idx_widget.setLayout(idx_layout)
        idx_widget.setMinimumHeight(55)
        idx_widget.setMinimumWidth(275)

        # rows
        rows = QVBoxLayout()
        # rows.addWidget(label)
        rows.addWidget(idx_widget)
        rows.addLayout(cols)  # raw option
        # rows.addWidget(self.window.ui.nodes['tip.toolbox.indexes'])

        self.window.ui.nodes['idx.options'] = QWidget()
        self.window.ui.nodes['idx.options'].setLayout(rows)
        self.window.ui.nodes['idx.options'].setContentsMargins(0, 0, 0, 0)

        return self.window.ui.nodes['idx.options']

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model

        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update(self, data):
        """
        Update list of indexes

        :param data: Data to update
        """
        # combo box
        combo_keys = []
        for item in data:
            name = item['name']
            if name == "":
                name = item['id']
            combo_keys.append({
                item['id']: name
            })
        self.window.ui.nodes['indexes.select'].set_keys(combo_keys)
        """
        # store previous selection
        self.window.ui.nodes[self.id].backup_selection()

        self.window.ui.models[self.id].removeRows(0, self.window.ui.models[self.id].rowCount())
        i = 0
        for item in data:
            self.window.ui.models[self.id].insertRow(i)
            name = item['name']
            index = self.window.ui.models[self.id].index(i, 0)
            self.window.ui.models[self.id].setData(index, "ID: " + item['id'], QtCore.Qt.ToolTipRole)
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 0), name)
            i += 1

        # restore previous selection
        self.window.ui.nodes[self.id].restore_selection()
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Flow:
    def __init__(self, window=None):
        """
        Agent flow controller

        :param window: Window instance
        """
        self.window = window
        self.iteration = 0
        self.prev_output = None
        self.is_user = True
        self.stop = False
        self.finished = False
        self.allowed_cmds = [
            "goal_update",
        ]
        self.pause_status = ["pause", "failed", "wait"]

    def on_system_prompt(
            self,
            prompt: str,
            append_prompt: str = "",
            auto_stop: bool = True,
    ) -> str:
        """
        Event: On prepare system prompt

        :param prompt: prompt
        :param append_prompt: extra prompt (instruction)
        :param auto_stop: auto stop
        :return: updated prompt
        """
        stop_cmd = ""
        if auto_stop:
            stop_cmd = "\n\n" + self.window.core.prompt.get("agent.goal")
        if append_prompt is not None and append_prompt.strip() != "":
            append_prompt = "\n" + append_prompt
        prompt += str(append_prompt) + stop_cmd
        return prompt

    def on_input_before(self, prompt: str) -> str:
        """
        Event: On user input before

        :param prompt: prompt
        :return: updated prompt
        """
        if not self.is_user:
            return prompt

        return "user: " + prompt  # add user prefix

    def on_user_send(self, text: str):
        """
        Event: On user send text

        :param text: text
        """
        self.iteration = 0
        self.prev_output = None
        self.is_user = True
        self.finished = False  # reset finished flag
        if self.stop:
            self.stop = False
        self.window.controller.agent.update()  # update status

    def on_ctx_end(
            self,
            ctx: CtxItem,
            iterations: int = 0,
    ):
        """
        Event: On context end

        :param ctx: CtxItem
        :param iterations: iterations
        """
        if self.stop:
            self.stop = False
            self.iteration = 0
            self.prev_output = None
            self.window.controller.agent.update()  # update status

        if iterations == 0 or self.iteration < int(iterations):
            self.iteration += 1
            self.window.controller.agent.update()  # update status
            if self.prev_output is not None and self.prev_output != "":
                self.window.controller.chat.input.send(
                    text=self.prev_output,
                    force=True,
                    internal=True,
                    prev_ctx=ctx,
                )
        # internal call will not trigger async mode and will hide the message from previous iteration
        elif self.iteration >= int(iterations):
            self.on_stop(auto=True)
            if self.window.core.config.get("agent.goal.notify"):
                self.window.ui.tray.show_msg(
                    trans("notify.agent.stop.title"),
                    trans("notify.agent.stop.content"),
                )

    def on_ctx_before(
            self,
            ctx: CtxItem,
            reverse_roles: bool = False,
    ):
        """
        Event: Before ctx

        :param ctx: CtxItem
        :param reverse_roles: reverse roles
        """
        ctx.internal = True  # always force internal call
        self.is_user = False
        if self.iteration == 0:
            ctx.first = True

        # reverse roles in ctx
        if self.iteration > 0 \
                and self.iteration % 2 != 0 \
                and reverse_roles:
            tmp_input_name = ctx.input_name
            tmp_output_name = ctx.output_name
            ctx.input_name = tmp_output_name
            ctx.output_name = tmp_input_name

    def on_ctx_after(self, ctx: CtxItem):
        """
        Event: After ctx

        :param ctx: CtxItem
        """
        self.prev_output = self.window.core.prompt.get("agent.continue")  # continue...
        if ctx.extra_ctx is not None and ctx.extra_ctx != "":
            self.prev_output = ctx.extra_ctx

    def on_cmd(
            self,
            ctx: CtxItem,
            cmds: list,
    ):
        """
        Event: On commands

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        if self.window.core.config.get('agent.auto_stop'):
            self.cmd(ctx, cmds)

    def cmd(
            self,
            ctx: CtxItem,
            cmds: list,
    ):
        """
        Event: On command

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        for item in my_commands:
            try:
                if item["cmd"] == "goal_update":
                    if item["params"]["status"] == "finished":
                        self.on_stop(auto=True)
                        self.window.ui.status(trans('status.finished'))  # show info
                        self.finished = True
                        if self.window.core.config.get("agent.goal.notify"):
                            self.window.ui.tray.show_msg(
                                trans("notify.agent.goal.title"),
                                trans("notify.agent.goal.content"),
                            )
                    elif item["params"]["status"] in self.pause_status:
                        self.on_stop(auto=True)
                        self.window.ui.status(trans('status.finished'))  # show info
                        self.finished = True
            except Exception as e:
                self.window.core.debug.error(e)
                return

    def on_stop(self, auto: bool = False):
        """
        Event: On force stop

        :param auto: auto
        """
        self.window.controller.chat.common.unlock_input()
        self.iteration = 0
        self.prev_output = None
        self.stop = True
        self.finished = False  # reset finished flag

        # update index if auto-index enabled
        if auto:
            self.window.controller.idx.on_ctx_end(
                ctx=None,
                mode="agent",
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu

from pygpt_net.utils import trans
import pygpt_net.icons_rc

from .lang import Lang
from .theme import Theme


class Config:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window
        self.lang = Lang(window)
        self.theme = Theme(window)

    def setup(self):
        """Setup config menu"""
        self.window.ui.menu['config.settings'] = QAction(QIcon(":/icons/settings_filled.svg"),
                                                         trans("menu.config.settings"), self.window)
        self.window.ui.menu['config.settings'].setMenuRole(QAction.MenuRole.NoRole)

        self.window.ui.menu['config.models'] = QAction(QIcon(":/icons/settings_filled.svg"),
                                                       trans("menu.config.models"), self.window)

        self.window.ui.menu['config.access'] = QAction(QIcon(":/icons/accessibility.svg"),
                                                       trans("menu.config.access"), self.window)

        css_files = []
        css_files.append("style.css")
        css_files.append("style.dark.css")
        css_files.append("style.light.css")
        css_files.append("markdown.css")
        css_files.append("markdown.dark.css")
        css_files.append("markdown.light.css")
        css_files.append("web.css")
        css_files.append("web.dark.css")
        css_files.append("web.light.css")

        json_files = []
        json_files.append("attachments.json")
        json_files.append("assistants.json")
        json_files.append("config.json")
        json_files.append("models.json")
        json_files.append("plugin_presets.json")

        # -------------------------------------------- #

        # create submenu for css files
        self.window.ui.menu['config.edit.css'] = QMenu(trans("menu.config.edit.css"), self.window)

        for css_file in css_files:
            name = css_file.split("/")[-1]
            self.window.ui.menu['config.edit.css.' + name] = QAction(QIcon(":/icons/edit.svg"),
                                                                     name, self.window)
            self.window.ui.menu['config.edit.css.' + name].triggered.connect(
                lambda checked=True, file=css_file: self.window.controller.settings.toggle_file_editor(file))
            self.window.ui.menu['config.edit.css'].addAction(self.window.ui.menu['config.edit.css.' + name])

        # restore css files
        self.window.ui.menu['config.edit.css'].addSeparator()
        self.window.ui.menu['config.edit.css.restore'] = QAction(QIcon(":/icons/undo.svg"),
                                                                 trans('menu.config.edit.css.restore'), self.window)
        self.window.ui.menu['config.edit.css.restore'].triggered.connect(
            lambda checked=True: self.window.controller.layout.restore_default_css(force=False))
        self.window.ui.menu['config.edit.css'].addAction(self.window.ui.menu['config.edit.css.restore'])

        # -------------------------------------------- #

        # create submenu for JSON files
        self.window.ui.menu['config.edit.json'] = QMenu(trans("menu.config.edit.json"), self.window)

        for json_file in json_files:
            name = json_file
            self.window.ui.menu['config.edit.json.' + name] = QAction(QIcon(":/icons/edit.svg"),
                                                                      name, self.window)
            self.window.ui.menu['config.edit.json.' + name].triggered.connect(
                lambda checked=True, file=json_file: self.window.controller.settings.toggle_file_editor(file))
            self.window.ui.menu['config.edit.json'].addAction(self.window.ui.menu['config.edit.json.' + name])

        # -------------------------------------------- #

        # create submenu for profiles
        self.window.ui.menu['config.profiles'] = {}
        self.window.ui.menu['config.profile'] = QMenu(trans("menu.config.profile"), self.window)

        # add new
        self.window.ui.menu['config.profile.new'] = QAction(
            QIcon(":/icons/add.svg"), trans("menu.config.profile.new"), self.window, checkable=False)
        self.window.ui.menu['config.profile.new'].triggered.connect(
            lambda: self.window.controller.settings.profile.new())
        self.window.ui.menu['config.profile.new'].setMenuRole(QAction.MenuRole.NoRole)
        self.window.ui.menu['config.profile'].addAction(self.window.ui.menu['config.profile.new'])

        # edit
        self.window.ui.menu['config.profile.edit'] = QAction(
            QIcon(":/icons/edit.svg"), trans("menu.config.profile.edit"), self.window, checkable=False)
        self.window.ui.menu['config.profile.edit'].triggered.connect(
            lambda: self.window.controller.settings.profile.toggle_editor())
        self.window.ui.menu['config.profile.edit'].setMenuRole(QAction.MenuRole.NoRole)
        self.window.ui.menu['config.profile'].addAction(self.window.ui.menu['config.profile.edit'])
        self.window.ui.menu['config.profile'].addSeparator()

        # -------------------------------------------- #

        self.window.ui.menu['config.open_dir'] = QAction(QIcon(":/icons/folder_filled.svg"),
                                                         trans("menu.config.open_dir"), self.window)
        self.window.ui.menu['config.change_dir'] = QAction(QIcon(":/icons/settings_filled.svg"),
                                                         trans("menu.config.change_dir"), self.window)
        self.window.ui.menu['config.save'] = QAction(QIcon(":/icons/save.svg"),
                                                     trans("menu.config.save"), self.window)

        self.window.ui.menu['config.settings'].triggered.connect(
            lambda: self.window.controller.settings.toggle_editor('settings'))

        self.window.ui.menu['config.models'].triggered.connect(
            lambda: self.window.controller.model.editor.toggle_editor())

        self.window.ui.menu['config.access'].triggered.connect(
            lambda: self.window.controller.settings.open_section('access'))

        self.window.ui.menu['config.open_dir'].triggered.connect(
            lambda: self.window.controller.settings.open_config_dir())

        self.window.ui.menu['config.change_dir'].triggered.connect(
            lambda: self.window.controller.settings.workdir.change())

        self.window.ui.menu['config.save'].triggered.connect(
            lambda: self.window.controller.settings.save_all())

        self.lang.setup()
        self.theme.setup()

        self.window.ui.menu['menu.config'] = self.window.menuBar().addMenu(trans("menu.config"))
        self.window.ui.menu['menu.config'].addAction(self.window.ui.menu['config.settings'])
        self.window.ui.menu['menu.config'].addAction(self.window.ui.menu['config.models'])
        self.window.ui.menu['menu.config'].addAction(self.window.ui.menu['config.access'])

        self.window.ui.menu['menu.config'].addMenu(self.window.ui.menu['menu.theme'])
        self.window.ui.menu['menu.config'].addMenu(self.window.ui.menu['menu.lang'])
        self.window.ui.menu['menu.config'].addMenu(self.window.ui.menu['config.edit.css'])
        self.window.ui.menu['menu.config'].addMenu(self.window.ui.menu['config.edit.json'])
        self.window.ui.menu['menu.config'].addMenu(self.window.ui.menu['config.profile'])
        self.window.ui.menu['menu.config'].addAction(self.window.ui.menu['config.open_dir'])
        self.window.ui.menu['menu.config'].addAction(self.window.ui.menu['config.change_dir'])
        self.window.ui.menu['menu.config'].addAction(self.window.ui.menu['config.save'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 17:00:00                  #
# ================================================== #

import json
import re

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem


class Chat:
    def __init__(self, window=None):
        """
        Chat wrapper

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0

    def send(self, context: BridgeContext, extra: dict = None):
        """
        Call OpenAI API for chat

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response or stream chunks
        """
        prompt = context.prompt
        stream = context.stream
        max_tokens = int(context.max_tokens or 0)
        system_prompt = context.system_prompt
        model = context.model
        functions = context.external_functions
        attachments = context.attachments

        ctx = context.ctx
        if ctx is None:
            ctx = CtxItem()  # create empty context
        user_name = ctx.input_name  # from ctx
        ai_name = ctx.output_name  # from ctx

        client = self.window.core.gpt.get_client()

        # build chat messages
        messages = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            history=context.history,
            attachments=attachments,
            ai_name=ai_name,
            user_name=user_name,
        )
        msg_tokens = self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        # check if max tokens not exceeded
        if max_tokens > 0:
            if msg_tokens + int(max_tokens) > model.ctx:
                max_tokens = model.ctx - msg_tokens - 1
                if max_tokens < 0:
                    max_tokens = 0

        # extra API kwargs
        response_kwargs = {}

        # tools / functions
        tools = []
        if functions is not None and isinstance(functions, list):
            for function in functions:
                if str(function['name']).strip() == '' or function['name'] is None:
                    continue
                params = json.loads(function['params'])  # unpack JSON from string
                tools.append(
                    {
                        "type": "function",
                        "function": {
                            "name": function['name'],
                            "parameters": params,
                            "description": function['desc'],
                        }
                    }
                )
        if len(tools) > 0:
            response_kwargs['tools'] = tools
        if max_tokens > 0:
            response_kwargs['max_tokens'] = max_tokens

        response = client.chat.completions.create(
            messages=messages,
            model=model.id,
            temperature=self.window.core.config.get('temperature'),
            top_p=self.window.core.config.get('top_p'),
            frequency_penalty=self.window.core.config.get('frequency_penalty'),
            presence_penalty=self.window.core.config.get('presence_penalty'),
            stream=stream,
            **response_kwargs
        )
        return response

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: list = None,
            attachments: dict = None,
            ai_name: str = None,
            user_name: str = None
    ) -> list:
        """
        Build list of chat messages

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param history: history
        :param model: model item
        :param attachments: attachments
        :param ai_name: AI name
        :param user_name: username
        :return: messages list
        """
        messages = []

        # tokens config
        mode = "chat"
        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )  # threshold and extra included
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')  # max context window

        # fit to max model tokens
        if max_ctx_tokens > model.ctx:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        # names fallback
        if ai_name is None or ai_name == "":
            ai_name = "assistant"
        if user_name is None or user_name == "":
            user_name = "user"

        # append system prompt
        if system_prompt is not None and system_prompt != "":
            messages.append({"role": "system", "content": system_prompt})

        # append messages from context (memory)
        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                mode,
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                # input
                if item.input is not None and item.input != "":
                    messages.append({"role": "system", "name": self.sanitize_name(user_name), "content": item.input})

                # output
                if item.output is not None and item.output != "":
                    messages.append({"role": "system", "name": self.sanitize_name(ai_name), "content": item.output})

        # use vision if available in current model
        content = str(prompt)
        if "vision" in model.mode:
            content = self.window.core.gpt.vision.build_content(prompt, attachments)

        # append current prompt
        messages.append({"role": "user", "content": content})

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        return messages

    def sanitize_name(self, name: str) -> str:
        """
        Sanitize name

        :param name: name
        :return: sanitized name
        """
        if name is None:
            return ""
        # allowed characters: a-z, A-Z, 0-9, _, and -
        name = name.strip().lower()
        sanitized_name = re.sub(r'[^a-z0-9_-]', '_', name)
        return sanitized_name[:64]  # limit to 64 characters

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

import pyaudio
import wave
import os

from PySide6.QtCore import QTimer

from pygpt_net.core.access.events import AppEvent
from pygpt_net.utils import trans


class Simple:

    TIMEOUT_SECONDS = 120  # 2 minutes, max recording time before timeout

    def __init__(self, plugin=None):
        """
        Simple audio input handler

        :param plugin: plugin instance
        """
        self.plugin = plugin
        self.is_recording = False
        self.frames = []
        self.p = None
        self.stream = None
        self.timer = None

    def toggle_recording(self):
        """Toggle recording"""
        if self.is_recording:
            self.stop_recording()
        else:
            self.start_recording()

    def switch_btn_stop(self):
        """Switch button to stop"""
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setText(trans('audio.speak.btn.stop'))
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setToolTip(
            trans('audio.speak.btn.stop.tooltip'))

    def switch_btn_start(self):
        """Switch button to start"""
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setText(trans('audio.speak.btn'))
        self.plugin.window.ui.plugin_addon['audio.input.btn'].btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))

    def stop_timeout(self):
        """Stop timeout"""
        self.stop_recording(timeout=True)

    def start_recording(self):
        """Start recording"""
        self.frames = []  # clear audio frames

        def callback(in_data, frame_count, time_info, status):
            self.frames.append(in_data)
            if self.is_recording:
                return (in_data, pyaudio.paContinue)
            else:
                return (in_data, pyaudio.paComplete)

        try:
            self.is_recording = True
            self.switch_btn_stop()

            # start timeout timer to prevent infinite recording
            if self.timer is None:
                self.timer = QTimer()
                self.timer.timeout.connect(self.stop_timeout)
                self.timer.start(self.TIMEOUT_SECONDS * 1000)

            self.p = pyaudio.PyAudio()
            self.stream = self.p.open(format=pyaudio.paInt16,
                                      channels=1,
                                      rate=44100,
                                      input=True,
                                      frames_per_buffer=1024,
                                      stream_callback=callback)

            self.plugin.window.ui.status(trans('audio.speak.now'))
            self.plugin.window.core.dispatcher.dispatch(AppEvent(AppEvent.INPUT_VOICE_LISTEN_STARTED))  # app event
            self.stream.start_stream()
        except Exception as e:
            self.is_recording = False
            self.plugin.window.core.debug.log(e)
            self.plugin.window.ui.dialogs.alert(e)
            if self.plugin.window.core.platforms.is_snap():
                self.plugin.window.ui.dialogs.open(
                    'snap_audio_input',
                    width=400,
                    height=200
                )
            self.switch_btn_start()  # switch button to start

    def stop_recording(self, timeout: bool = False):
        """
        Stop recording

        :param timeout: True if stopped due to timeout
        """
        self.is_recording = False
        self.plugin.window.core.dispatcher.dispatch(AppEvent(AppEvent.INPUT_VOICE_LISTEN_STOPPED))  # app event
        if self.timer:
            self.timer.stop()
            self.timer = None
        self.switch_btn_start()  # switch button to start
        path = os.path.join(self.plugin.window.core.config.path, self.plugin.input_file)

        if self.stream is not None:
            self.stream.stop_stream()
            self.stream.close()
            self.p.terminate()

            # abort if timeout
            if timeout:
                self.plugin.window.ui.status("Aborted.".format(self.TIMEOUT_SECONDS))
                return

            if self.frames:
                wf = wave.open(path, 'wb')
                wf.setnchannels(1)
                wf.setsampwidth(self.p.get_sample_size(pyaudio.paInt16))
                wf.setframerate(44100)
                wf.writeframes(b''.join(self.frames))
                wf.close()
                self.plugin.handle_thread(True)  # handle transcription in simple mode



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLabel, QHBoxLayout, QWidget, QPushButton

from pygpt_net.core.access.events import AppEvent
from pygpt_net.core.dispatcher import Event
from pygpt_net.utils import trans

class VoiceControlButton(QWidget):
    def __init__(self, window=None):
        """
        VoiceControl button

        :param window: Window instance
        """
        super(VoiceControlButton, self).__init__(window)
        self.window = window

        self.btn_toggle = QPushButton(trans('audio.control.btn'))
        self.btn_toggle.clicked.connect(self.toggle_recording)
        self.btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))
        self.btn_toggle.setCursor(Qt.PointingHandCursor)
        self.btn_toggle.setMinimumWidth(200)

        # status
        self.status = QLabel("")
        self.status.setStyleSheet("color: #999; font-size: 10px; font-weight: 400; margin: 0; padding: 0; border: 0;")
        self.status.setMaximumHeight(15)

        self.layout = QHBoxLayout(self)
        self.layout.addWidget(self.btn_toggle)
        self.layout.addWidget(self.status)

        # self.layout.addWidget(self.stop)
        self.layout.setAlignment(Qt.AlignCenter)
        self.setLayout(self.layout)
        self.setMaximumHeight(80)

    def add_widget(self, widget):
        """
        Add widget to layout

        :param widget: widget
        """
        self.layout.addWidget(widget)

    def set_status(self, text):
        """
        Set status text

        :param text: text
        """
        self.status.setText(text)

    def toggle_recording(self):
        """Toggle voice control"""
        self.window.core.dispatcher.dispatch(AppEvent(AppEvent.VOICE_CONTROL_TOGGLE))  # app event


class AudioInputButton(QWidget):
    def __init__(self, window=None):
        """
        Audio input record button

        :param window: Window instance
        """
        super(AudioInputButton, self).__init__(window)
        self.window = window

        self.btn_toggle = QPushButton(trans('audio.speak.btn'))
        self.btn_toggle.clicked.connect(self.toggle_recording)
        self.btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))
        self.btn_toggle.setCursor(Qt.PointingHandCursor)
        self.btn_toggle.setMinimumWidth(200)

        # status
        self.status = QLabel("")
        self.status.setStyleSheet("color: #999; font-size: 10px; font-weight: 400; margin: 0; padding: 0; border: 0;")
        self.status.setMaximumHeight(15)

        self.layout = QHBoxLayout(self)
        self.layout.addWidget(self.btn_toggle)
        self.layout.addWidget(self.status)

        # self.layout.addWidget(self.stop)
        self.layout.setAlignment(Qt.AlignCenter)
        self.setLayout(self.layout)
        self.setMaximumHeight(80)

    def add_widget(self, widget):
        """
        Add widget to layout

        :param widget: widget
        """
        self.layout.addWidget(widget)

    def set_status(self, text):
        """
        Set status text

        :param text: text
        """
        self.status.setText(text)

    def toggle_recording(self):
        """Toggle recording"""
        event = Event(Event.AUDIO_INPUT_RECORD_TOGGLE)
        self.window.core.dispatcher.dispatch(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.17 01:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QPushButton, QLabel, QVBoxLayout, QWidget, QHBoxLayout

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans


class CtxTab:
    def __init__(self, window=None):
        """
        Tab: Context indexing

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """
        Setup tab widget
        """
        self.window.ui.nodes["tool.indexer.ctx.last_auto.label"] = HelpLabel(trans("tool.indexer.tab.ctx.last_auto"))
        self.window.ui.nodes["tool.indexer.ctx.last_auto"] = QLabel("")

        self.window.ui.nodes["tool.indexer.ctx.auto_enabled.label"] = HelpLabel(trans("tool.indexer.tab.ctx.auto_enabled"))
        self.window.ui.nodes["tool.indexer.ctx.auto_enabled"] = QLabel("")

        self.window.ui.nodes["tool.indexer.ctx.last_meta_id.label"] = HelpLabel(trans("tool.indexer.tab.ctx.last_meta_id"))
        self.window.ui.nodes["tool.indexer.ctx.last_meta_id"] = QLabel("")

        self.window.ui.nodes["tool.indexer.ctx.last_meta_ts.label"] = HelpLabel(trans("tool.indexer.tab.ctx.last_meta_ts"))
        self.window.ui.nodes["tool.indexer.ctx.last_meta_ts"] = QLabel("")

        # auto db index enabled
        auto_enabled_layout = QHBoxLayout()
        auto_enabled_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.auto_enabled.label"])
        auto_enabled_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.auto_enabled"])

        # auto db index TS
        last_auto_layout = QHBoxLayout()
        last_auto_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.last_auto.label"])
        last_auto_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.last_auto"])

        # last meta in DB
        last_meta_layout = QHBoxLayout()
        last_meta_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.last_meta_id.label"])
        last_meta_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.last_meta_id"])

        last_meta_ts_layout = QHBoxLayout()
        last_meta_ts_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.last_meta_ts.label"])
        last_meta_ts_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.last_meta_ts"])

        self.window.ui.nodes["tool.indexer.ctx.btn.idx_all"] = QPushButton(trans('settings.llama.extra.btn.idx_db_all'))
        self.window.ui.nodes["tool.indexer.ctx.btn.idx_all"].clicked.connect(
            lambda: self.window.tools.get("indexer").idx_ctx_db_all()
        )
        self.window.ui.nodes["tool.indexer.ctx.btn.idx_update"] = QPushButton(
            trans('settings.llama.extra.btn.idx_db_update'))
        self.window.ui.nodes["tool.indexer.ctx.btn.idx_update"].clicked.connect(
            lambda: self.window.tools.get("indexer").idx_ctx_db_update()
        )

        idx_btn_layout = QHBoxLayout()
        idx_btn_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.btn.idx_all"])
        idx_btn_layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.btn.idx_update"])

        self.window.ui.nodes["tool.indexer.ctx.idx.tip"] = QLabel(trans("tool.indexer.tab.ctx.idx.tip"))
        self.window.ui.nodes["tool.indexer.ctx.header.tip"] = HelpLabel(trans("tool.indexer.tab.ctx.tip"))

        # layout
        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.header.tip"])
        layout.addLayout(auto_enabled_layout)
        layout.addLayout(last_auto_layout)
        layout.addLayout(last_meta_layout)
        layout.addLayout(last_meta_ts_layout)
        layout.addWidget(QLabel(""))
        layout.addWidget(self.window.ui.nodes["tool.indexer.ctx.idx.tip"])
        layout.addLayout(idx_btn_layout)
        layout.addStretch(1)

        self.window.tools.get("indexer").update_tab_ctx()

        widget = QWidget()
        widget.setLayout(layout)

        return widget


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #
import os

from pygpt_net.core.dispatcher import Event
from pygpt_net.plugin.audio_output.worker import PlayWorker
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Audio:
    def __init__(self, window=None):
        """
        Audio/voice controller

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup controller"""
        self.update()

    def toggle_input(self, state: bool, btn: bool = True):
        """
        Toggle audio input

        :param state: True to enable, False to disable
        :param btn: True if called from button
        """
        self.window.core.dispatcher.dispatch(
            Event(Event.AUDIO_INPUT_TOGGLE, {
                "value": state,
            })
        )

    def toggle_output(self):
        """Toggle audio output"""
        if self.window.controller.plugins.is_enabled('audio_output'):
            self.disable_output()
        else:
            self.enable_output()

    def enable_output(self):
        """Enable audio output"""
        self.toggle_output_icon(True)
        self.window.controller.plugins.enable('audio_output')
        self.window.core.config.save()
        self.update()

    def disable_output(self):
        """Disable audio output"""
        self.toggle_output_icon(False)
        self.window.controller.plugins.disable('audio_output')
        self.window.core.config.save()
        self.update()

    def disable_input(self, update: bool = True):
        """
        Disable audio input

        :param update: True to update menu and listeners
        """
        self.window.controller.plugins.disable('audio_input')
        self.window.core.config.save()
        if update:
            self.update()

    def stop_input(self):
        """Stop audio input"""
        self.window.core.dispatcher.dispatch(
            Event(Event.AUDIO_INPUT_STOP, {
                "value": True,
            }), all=True)

    def stop_output(self):
        """Stop audio output"""
        self.window.core.dispatcher.dispatch(
            Event(Event.AUDIO_OUTPUT_STOP, {
                "value": True,
            }), all=True)

    def update(self):
        """Update UI and listeners"""
        self.update_listeners()
        self.update_menu()

    def is_output_enabled(self) -> bool:
        """
        Check if any audio output is enabled

        :return: True if enabled
        """
        if self.window.controller.plugins.is_enabled('audio_output'):
            return True
        return False

    def update_listeners(self):
        """Update audio listeners"""
        is_output = False
        if self.window.controller.plugins.is_enabled('audio_output'):
            is_output = True
        if not is_output:
            self.stop_output()

        if not self.window.controller.plugins.is_enabled('audio_input'):
            self.toggle_input(False)
            self.stop_input()
            if self.window.ui.plugin_addon['audio.input'].btn_toggle.isChecked():
                self.window.ui.plugin_addon['audio.input'].btn_toggle.setChecked(False)

    def update_menu(self):
        """Update audio menu"""
        if self.window.controller.plugins.is_enabled('audio_output'):
            self.window.ui.menu['audio.output'].setChecked(True)
        else:
            self.window.ui.menu['audio.output'].setChecked(False)

        if self.window.controller.plugins.is_enabled('audio_input'):
            self.window.ui.menu['audio.input'].setChecked(True)
        else:
            self.window.ui.menu['audio.input'].setChecked(False)

    def read_text(self, text: str):
        """
        Read text using audio output plugins

        :param text: text to read
        """
        ctx = CtxItem()
        ctx.output = text
        all = True  # to all plugins (even if disabled)
        event = Event(Event.AUDIO_READ_TEXT)
        event.ctx = ctx
        self.window.core.dispatcher.dispatch(event, all=all)

    def play_audio(self, path: str):
        """
        Play audio file

        :param path: audio file path
        """
        worker = PlayWorker()
        worker.window = self.window
        worker.path = path
        self.window.threadpool.start(worker)

    def play_sound(self, filename: str):
        """
        Play sound

        :param filename
        """
        path = os.path.join(self.window.core.config.get_app_path(), "data", "audio", filename)
        if path:
            self.play_audio(path)

    def toggle_output_icon(self, state: bool):
        """
        Toggle input icon

        :param state: True to enable, False to disable
        """
        if state:
            self.window.ui.nodes['icon.audio.output'].set_icon(":/icons/volume.svg")
        else:
            self.window.ui.nodes['icon.audio.output'].set_icon(":/icons/mute.svg")

    def toggle_input_icon(self, state: bool):
        """
        Toggle input icon

        :param state: True to enable, False to disable
        """
        if state:
            self.window.ui.nodes['icon.audio.input'].set_icon(":/icons/mic.svg")
        else:
            self.window.ui.nodes['icon.audio.input'].set_icon(":/icons/mic_off.svg")

    def on_begin(self, text: str):
        """
        On audio playback init

        :param text: text to play
        """
        self.window.ui.status(trans("status.audio.start"))

    def on_play(self, event: str):
        """
        On audio playback start

        :param event: event name
        """
        if event == Event.AUDIO_READ_TEXT:
            self.window.ui.status("")

    def on_stop(self):
        """
        On audio playback stopped (force)
        """
        self.window.ui.status(trans("status.audio.stopped"))

    def is_playing(self) -> bool:
        """
        Check if any audio is playing

        :return: True if playing
        """
        from pygame import mixer
        try:
            mixer.init()
            return mixer.get_busy()
        except Exception as e:
            pass
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

class Agents:
    def __init__(self, window=None):
        """
        Agents core

        :param window: Window instance
        """
        self.window = window
        self.allowed_modes = ["chat", "completion", "vision", "langchain", "llama_index"]

    def get_allowed_modes(self) -> list:
        """
        Get allowed modes

        :return: allowed modes
        """
        return self.allowed_modes

    def get_mode(self) -> str:
        """
        Get sub-mode to use internally

        :return: sub-mode
        """
        mode = "chat"
        current = self.window.core.config.get("agent.mode")
        if current is not None and current in self.allowed_modes:
            mode = current
        return mode

    def get_idx(self) -> str:
        """
        Get agent index

        :return: agent index
        """
        return self.window.core.config.get("agent.idx")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 07:00:00                  #
# ================================================== #

import datetime

from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans
from .common import Common
from .indexer import Indexer
from .settings import Settings


class Idx:
    def __init__(self, window=None):
        """
        Indexes controller

        :param window: Window instance
        """
        self.window = window
        self.settings = Settings(window)
        self.common = Common(window)
        self.indexer = Indexer(window)
        self.current_idx = "base"
        self.locked = False

    def setup(self):
        """Setup indexer"""
        self.window.core.idx.load()
        self.indexer.update_explorer()
        self.common.setup()

        # restore last index
        last_idx = self.window.core.config.get('llama.idx.current')
        if last_idx is not None:
            self.current_idx = last_idx

        self.locked = True  # lock update from combo box on start
        self.update()
        self.locked = False

    def select(self, idx: int):
        """
        Select idx by list idx

        :param idx: idx of the list (row idx)
        """
        # check if idx change is not locked
        if self.change_locked():
            return
        self.set_by_idx(idx)

        # update all layout
        self.window.controller.ui.update()

    def select_by_id(self, id: int):
        """
        Select idx by list idx

        :param id: id of the list (row idx)
        """
        # check if idx change is not locked
        if self.change_locked():
            return

        self.window.core.config.set('llama.idx.current', id)
        self.current_idx = id

        # update all layout
        self.window.controller.ui.update()

    def set(self, idx: str):
        """
        Set idx by name

        :param idx: idx name
        """
        self.window.core.config.set('llama.idx.current', idx)
        self.current_idx = idx

    def idx_db_update_by_idx(self, idx: int):
        """
        Index new records in database (update)

        :param idx: idx of the list (row idx)
        """
        idx = self.window.core.idx.get_by_idx(idx)
        if idx is None:
            return
        self.indexer.index_ctx_current(idx)

    def idx_db_all_by_idx(self, idx: int):
        """
        Index all records in database

        :param idx: idx of the list (row idx)
        """
        idx = self.window.core.idx.get_by_idx(idx)
        if idx is None:
            return
        self.indexer.index_ctx_from_ts(idx, 0)

    def idx_files_all_by_idx(self, idx: int):
        """
        Index all files in database

        :param idx: idx of the list (row idx)
        """
        idx = self.window.core.idx.get_by_idx(idx)
        if idx is None:
            return
        self.indexer.index_all_files(idx)

    def set_by_idx(self, idx: int):
        """
        Set idx by list idx

        :param idx: idx of the list (row idx)
        """
        idx = self.window.core.idx.get_by_idx(idx)
        if idx is None:
            return
        self.window.core.config.set('llama.idx.current', idx)
        self.current_idx = idx

    def select_current(self):
        """Select current idx on list"""
        idx = self.window.core.config.get('llama.idx.current')
        if idx is None:
            return
        items = self.window.core.config.get('llama.idx.list')
        if items is not None:
            self.window.ui.nodes['indexes.select'].set_value(idx)

    def select_default(self):
        """Set default idx"""
        idx = self.window.core.config.get('llama.idx.current')
        if idx is None:
            idx = self.window.core.idx.get_default_idx()
            if idx is not None:
                self.current_idx = idx

    def update(self):
        """Update idx list"""
        self.select_default()
        self.update_list()
        self.select_current()

    def update_list(self):
        """Update list"""
        items = self.window.core.config.get('llama.idx.list')
        if items is not None:
            self.window.ui.toolbox.indexes.update(items)

    def auto_idx_allowed(self, mode: str) -> bool:
        """
        Check if auto idx is allowed

        :param mode: mode name
        :return: True if allowed
        """
        modes = self.window.core.config.get('llama.idx.auto.modes')
        if modes is not None:
            modes_list = modes.replace(" ", "").split(',')
            if mode in modes_list:
                return True
        return False

    def on_ctx_end(self, ctx: CtxItem = None, mode: str = None, sync: bool = False):
        """
        After context item updated (request + response received)

        :param ctx: Context item instance
        :param mode: Mode
        :param sync: Synchronous call
        """
        # ignore if disallowed mode
        if mode is not None and not self.auto_idx_allowed(mode):
            return

        # ignore if manually stopped
        if self.window.controller.chat.input.stop:
            return

        idx = "base"  # default index
        if self.window.core.config.has('llama.idx.auto') and self.window.core.config.get('llama.idx.auto'):
            if self.window.core.config.has('llama.idx.auto.index'):
                idx = self.window.core.config.get('llama.idx.auto.index')

            # index items from previously indexed time only
            current_ctx = self.window.core.ctx.current
            if current_ctx is not None:
                meta = self.window.core.ctx.get_meta_by_id(current_ctx)
                if meta is not None:
                    self.indexer.index_ctx_realtime(meta, idx, sync=sync)

    def after_index(self, idx: str = None):
        """
        Called after index (update things, etc...)

        :param idx: index name
        """
        self.indexer.update_explorer()  # update file explorer view

        # update last indexing timestamp label
        last_str = '---'
        if self.window.core.config.has('llama.idx.db.last'):
            last_ts = int(self.window.core.config.get('llama.idx.db.last'))
            if last_ts > 0:
                last_str = datetime.datetime.fromtimestamp(last_ts).strftime('%Y-%m-%d %H:%M:%S')

        txt = trans('idx.last') + ": " + last_str
        self.window.ui.nodes['idx.db.last_updated'].setText(txt)

    def refresh(self):
        """Update list"""
        self.select_default()

    def change_locked(self) -> bool:
        """
        Check if change is locked

        :return: True if locked
        """
        return self.locked

    def reload(self):
        """Reload indexer"""
        self.setup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.24 02:00:00                  #
# ================================================== #

from pygments.styles import get_all_styles
from PySide6.QtGui import QAction


class Menu:
    def __init__(self, window=None):
        """
        Theme menu controller

        :param window: Window instance
        """
        self.window = window
        self.density_values = [-2, -1, 0, 1, 2]
        self.loaded = False
        self.syntax_loaded = False
        self.density_loaded = False

    def setup_list(self):
        """Setup menu list"""
        # setup themes list menu
        if self.loaded:
            return
        themes = self.window.controller.theme.common.get_themes_list()
        for theme in themes:
            name = self.window.controller.theme.common.translate(theme)
            self.window.ui.menu['theme'][theme] = QAction(name, self.window, checkable=True)
            self.window.ui.menu['theme'][theme].triggered.connect(
                lambda checked=None, theme=theme: self.window.controller.theme.toggle(theme))

            # append to dark or light menu
            if theme.startswith('dark'):
                self.window.ui.menu['theme.dark'].addAction(self.window.ui.menu['theme'][theme])
            elif theme.startswith('light'):
                self.window.ui.menu['theme.light'].addAction(self.window.ui.menu['theme'][theme])
        self.loaded = True

    def setup_syntax(self):
        """Setup syntax menu"""
        styles = list(get_all_styles())
        styles.sort()
        # clear menu
        for style in self.window.ui.menu['theme_syntax']:
            self.window.ui.menu['theme.syntax'].removeAction(self.window.ui.menu['theme_syntax'][style])
        # setup syntax menu
        for style in styles:
            self.window.ui.menu['theme_syntax'][style] = QAction(style, self.window, checkable=True)
            self.window.ui.menu['theme_syntax'][style].triggered.connect(
                lambda checked=None, style=style: self.window.controller.theme.toggle_syntax(style, update_menu=True))
            self.window.ui.menu['theme.syntax'].addAction(self.window.ui.menu['theme_syntax'][style])

    def setup_density(self):
        """Setup menu list"""
        if self.density_loaded:
            return
        # setup layout density menu
        current_density = self.window.core.config.get('layout.density')
        for value in self.density_values:
            name = str(value)
            if value > 0:
                name = '+' + name
            self.window.ui.menu['theme.layout.density'][value] = QAction(name, self.window, checkable=True)
            self.window.ui.menu['theme.layout.density'][value].triggered.connect(
                lambda checked=None, value=value: self.window.controller.theme.toggle_option('layout.density', value))
            self.window.ui.menu['theme.density'].addAction(self.window.ui.menu['theme.layout.density'][value])
            if value == current_density:
                self.window.ui.menu['theme.layout.density'][value].setChecked(True)
        self.density_loaded = True

    def update_density(self):
        """Update layout density menu"""
        current_density = self.window.core.config.get('layout.density')
        for value in self.density_values:
            self.window.ui.menu['theme.layout.density'][value].setChecked(False)
            if value == current_density:
                self.window.ui.menu['theme.layout.density'][value].setChecked(True)

    def update_list(self):
        """Update theme list menu"""
        current = self.window.core.config.get('theme')
        for theme in self.window.ui.menu['theme']:
            self.window.ui.menu['theme'][theme].setChecked(False)
        if current in self.window.ui.menu['theme']:
            self.window.ui.menu['theme'][current].setChecked(True)

    def update_syntax(self):
        """Update syntax menu"""
        current = self.window.core.config.get('render.code_syntax')
        for style in self.window.ui.menu['theme_syntax']:
            self.window.ui.menu['theme_syntax'][style].setChecked(False)
        if current in self.window.ui.menu['theme_syntax']:
            self.window.ui.menu['theme_syntax'][current].setChecked(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 17:00:00                  #
# ================================================== #

import json
import uuid


class PresetItem:
    def __init__(self):
        self.uuid = None
        self.name = "*"
        self.ai_name = ""
        self.user_name = ""
        self.prompt = ""
        self.chat = False
        self.completion = False
        self.img = False
        self.vision = False
        self.langchain = False
        self.assistant = False
        self.llama_index = False
        self.agent = False
        self.expert = False
        self.temperature = 1.0
        self.filename = None
        self.model = None
        self.version = None
        self.experts = []  # agent mode
        self.enabled = True
        self.tools = {
            "function": [],
        }

    def get_id(self) -> str:
        return self.filename

    def to_dict(self):
        return {
            "uuid": str(self.uuid),
            "name": self.name,
            "ai_name": self.ai_name,
            "user_name": self.user_name,
            "prompt": self.prompt,
            "chat": self.chat,
            "completion": self.completion,
            "img": self.img,
            "vision": self.vision,
            "langchain": self.langchain,
            "assistant": self.assistant,
            "llama_index": self.llama_index,
            "agent": self.agent,
            "expert": self.expert,
            "temperature": self.temperature,
            "filename": self.filename,
            "model": self.model,
            "tool.function": self.tools["function"],
            "experts": self.experts,
            "enabled": self.enabled,
        }

    def from_dict(self, data):
        if "uuid" in data:
            self.uuid = uuid.UUID(data["uuid"])
        if "name" in data:
            self.name = data["name"]
        if "ai_name" in data:
            self.ai_name = data["ai_name"]
        if "user_name" in data:
            self.user_name = data["user_name"]
        if "prompt" in data:
            self.prompt = data["prompt"]
        if "chat" in data:
            self.chat = data["chat"]
        if "completion" in data:
            self.completion = data["completion"]
        if "img" in data:
            self.img = data["img"]
        if "vision" in data:
            self.vision = data["vision"]
        if "langchain" in data:
            self.langchain = data["langchain"]
        if "assistant" in data:
            self.assistant = data["assistant"]
        if "llama_index" in data:
            self.llama_index = data["llama_index"]
        if "agent" in data:
            self.agent = data["agent"]
        if "expert" in data:
            self.expert = data["expert"]
        if "temperature" in data:
            self.temperature = data["temperature"]
        if "filename" in data:
            self.filename = data["filename"]
        if "model" in data:
            self.model = data["model"]
        if "tool.function" in data:
            self.tools["function"] = data["tool.function"]
        if "experts" in data:
            self.experts = data["experts"]
        if "enabled" in data:
            self.enabled = data["enabled"]
        return self

    def add_function(self, name: str, parameters: str, desc: str):
        """
        Add function to preset

        :param name: function name
        :param parameters: function parameters (JSON encoded)
        :param desc: function description
        """
        function = {
            'name': name,
            'params': parameters,
            'desc': desc,
        }
        self.tools['function'].append(function)

    def has_functions(self) -> bool:
        """
        Check if preset has functions

        :return: bool
        """
        return len(self.tools['function']) > 0

    def get_functions(self) -> list:
        """
        Return preset functions

        :return: functions
        """
        return self.tools['function']

    def dump(self):
        """
        Dump item to string

        :return: serialized item
        :rtype: str
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 07:00:00                  #
# ================================================== #

import os

from .editor import Editor
from .profile import Profile
from .workdir import Workdir

from pygpt_net.utils import trans


class Settings:
    def __init__(self, window=None):
        """
        Settings controller

        :param window: Window instance
        """
        self.window = window
        self.editor = Editor(window)
        self.profile = Profile(window)
        self.workdir = Workdir(window)
        self.width = 800
        self.height = 500

    def setup(self):
        """Set up settings editor"""
        self.profile.setup()
        self.editor.setup()

    def load(self):
        """Load settings"""
        self.editor.load()

    def save_all(self, force: bool = False):
        """
        Save all settings and data

        :param force: force save
        """
        info = trans('info.settings.all.saved')
        self.window.core.config.save()
        self.window.core.presets.save_all()
        self.window.controller.notepad.save_all()
        self.window.controller.calendar.save_all()
        self.window.controller.painter.save()
        if not force:
            self.window.ui.dialogs.alert(info)
            self.window.ui.status(info)
        self.window.controller.ui.update()

    def update(self):
        """Update settings"""
        self.update_menu()

    def update_menu(self):
        """Update menu"""
        for id in self.window.core.settings.ids:
            key = 'config.' + id
            if key in self.window.ui.menu:
                if id in self.window.core.settings.active and self.window.core.settings.active[id]:
                    self.window.ui.menu['config.' + id].setChecked(True)
                else:
                    self.window.ui.menu['config.' + id].setChecked(False)

    def open_section(self, section: str):
        """
        Open settings section

        :param section: section id
        """
        id = 'settings'
        tab = self.get_tab_idx(section)
        if tab is not None:
            self.set_by_tab(tab)

        self.window.ui.dialogs.open(
            'config.' + id,
            width=self.width,
            height=self.height,
        )
        self.editor.init(id)
        self.window.core.settings.active[id] = True
        self.window.controller.layout.restore_settings()

        # update menu
        self.update()

    def toggle_editor(self, id: str):
        """
        Toggle settings dialog

        :param id: settings id
        """
        if id in self.window.core.settings.active and self.window.core.settings.active[id]:
            self.close_window(id)
        else:
            self.window.ui.dialogs.open('config.' + id, width=self.width, height=self.height)
            self.editor.init(id)
            self.window.core.settings.active[id] = True

            # if no API key, focus on API key input
            if self.window.core.config.get('api_key') is None or self.window.core.config.get('api_key') == '':
                self.window.ui.config['config']['api_key'].setFocus()

            self.window.controller.layout.restore_settings()  # restore previous selected settings tab

        # update menu
        self.update()

    def set_by_tab(self, idx: int):
        """
        Set current section by tab index

        :param idx: tab index
        """
        section_idx = 0
        for section_id in self.editor.get_sections():
            if section_idx == idx:
                break
            section_idx += 1
        current = self.window.ui.models['settings.section.list'].index(idx, 0)
        self.window.ui.nodes['settings.section.list'].setCurrentIndex(current)
        self.window.ui.tabs['settings.section'].setCurrentIndex(idx)

    def get_tab_idx(self, section_id: str) -> int:
        """
        Get section tab index

        :param section_id: plugin id
        :return: tab index
        """
        section_idx = None
        i = 0
        for id in self.editor.get_sections():
            if id == section_id:
                section_idx = i
                break
            i += 1
        return section_idx

    def toggle_file_editor(self, file: str = None):
        """
        Toggle file editor

        :param file: JSON file to load
        """
        id = 'editor'
        current_file = self.window.ui.dialog['config.editor'].file

        # show/hide restart required label
        if file.endswith('.css'):
            self.window.ui.nodes['dialog.editor.label'].setVisible(False)
        else:
            self.window.ui.nodes['dialog.editor.label'].setVisible(True)

        if id in self.window.core.settings.active and self.window.core.settings.active[id]:
            if current_file == file:
                self.window.ui.dialogs.close('config.' + id)
                self.window.core.settings.active[id] = False
            else:
                self.window.core.settings.load_editor(file)  # load file to editor
                self.window.ui.dialog['config.editor'].file = file
                self.prepare_file_editor(file)
        else:
            self.window.core.settings.load_editor(file)  # load file to editor
            self.prepare_file_editor(file)
            self.window.ui.dialogs.open(
                'config.' + id,
                width=self.width,
                height=self.height,
            )
            self.window.core.settings.active[id] = True

        # update menu
        self.update()

    def prepare_file_editor(self, file: str):
        """
        Prepare file editor

        :param file: JSON/CSS file to load
        """
        return  # allow defaults to all files
        if file.endswith('.css'):
            self.window.ui.nodes['editor.btn.default'].setVisible(True)
        else:
            self.window.ui.nodes['editor.btn.default'].setVisible(False)

    def close(self, id: str):
        """
        Close menu

        :param id: settings window id
        """
        if id in self.window.ui.menu:
            self.window.ui.menu[id].setChecked(False)

        allowed_settings = ['settings']
        if id in allowed_settings and id in self.window.ui.menu:
            self.window.ui.menu[id].setChecked(False)

    def close_window(self, id: str):
        """
        Close window

        :param id: settings window id
        """
        if id in self.window.core.settings.active and self.window.core.settings.active[id]:
            self.window.ui.dialogs.close('config.' + id)
            self.window.core.settings.active[id] = False

    def open_config_dir(self):
        """Open user config directory"""
        if os.path.exists(self.window.core.config.path):
            self.window.controller.files.open_dir(self.window.core.config.path)
        else:
            self.window.ui.status('Config directory not exists: {}'.format(self.window.core.config.path))

    def welcome_settings(self):
        """Open settings at first launch (if no API key yet)"""
        self.open_section("general")
        self.window.ui.config['config']['api_key'].setFocus()
        self.window.ui.dialogs.close('info.start')

    def reload(self):
        """Reload settings"""
        self.setup()
        self.window.ui.dialogs.app_log.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.30 15:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal

from pygpt_net.plugin.base import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    updated = Signal()  # calendar view update signal


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        msg = None
        for item in self.cmds:
            response = None
            try:
                if item["cmd"] == "get_ctx_list_in_date_range":
                    response = self.cmd_get_ctx_list_in_date_range(item)

                elif item["cmd"] == "get_ctx_content_by_id":
                    response = self.cmd_get_ctx_content_by_id(item)

                elif item["cmd"] == "get_day_note":
                    response = self.cmd_get_day_note(item)

                elif item["cmd"] == "add_day_note":
                    response = self.cmd_add_day_note(item)

                elif item["cmd"] == "update_day_note":
                    response = self.cmd_update_day_note(item)

                elif item["cmd"] == "remove_day_note":
                    response = self.cmd_remove_day_note(item)

                elif item["cmd"] == "count_ctx_in_date":
                    response = self.cmd_count_ctx_in_date(item)

                if response:
                    responses.append(response)

            except Exception as e:
                msg = "Error: {}".format(e)
                responses.append({
                    "request": {
                        "cmd": item["cmd"],
                    },
                    "result": "Error {}".format(e),
                })
                self.error(e)
                self.log(msg)

        # send response
        if len(responses) > 0:
            for response in responses:
                self.reply(response)

        # update status
        if msg is not None:
            self.status(msg)

    def cmd_get_ctx_list_in_date_range(self, item: dict) -> dict:
        """
        Get context list in date range

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        range = item["params"]["range_query"]
        response = {
            "request": request,
            "result":  self.plugin.get_list(range)
        }
        return response

    def cmd_get_ctx_content_by_id(self, item: dict) -> dict:
        """
        Get context content by id

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        id = int(item["params"]["id"])
        prompt = "Summary this conversation"
        if "summary_query" in item["params"]:
            prompt = item["params"]["summary_query"]
        response = {
            "request": request,
            "result": self.plugin.get_summary(id, prompt),
        }
        return response

    def cmd_get_day_note(self, item: dict) -> dict:
        """
        Get day note

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        year = int(item["params"]["year"])
        month = int(item["params"]["month"])
        day = int(item["params"]["day"])
        response = {
            "request": request,
            "result": self.plugin.get_day_note(year, month, day),
        }
        return response

    def cmd_add_day_note(self, item: dict) -> dict:
        """
        Add day note

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        year = int(item["params"]["year"])
        month = int(item["params"]["month"])
        day = int(item["params"]["day"])
        note = item["params"]["note"]
        response = {
            "request": request,
            "result": self.plugin.add_day_note(year, month, day, note),
        }
        self.signals.updated.emit()
        return response

    def cmd_update_day_note(self, item: dict) -> dict:
        """
        Update day note

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        year = int(item["params"]["year"])
        month = int(item["params"]["month"])
        day = int(item["params"]["day"])
        note = item["params"]["content"]
        response = {
            "request": request,
            "result": self.plugin.update_day_note(year, month, day, note),
        }
        self.signals.updated.emit()
        return response

    def cmd_remove_day_note(self, item: dict) -> dict:
        """
        Remove day note

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        year = int(item["params"]["year"])
        month = int(item["params"]["month"])
        day = int(item["params"]["day"])
        response = {
            "request": request,
            "result": self.plugin.remove_day_note(year, month, day),
        }
        self.signals.updated.emit()
        return response

    def cmd_count_ctx_in_date(self, item: dict) -> dict:
        """
        Count context in date

        :param item: command item
        :return: response item
        """
        request = self.prepare_request(item)
        year = None
        month = None
        day = None
        if "year" in item["params"] and item["params"]["year"] != "":
            year = int(item["params"]["year"])
        if "month" in item["params"] and item["params"]["month"] != "":
            month = int(item["params"]["month"])
        if "day" in item["params"] and item["params"]["day"] != "":
            day = int(item["params"]["day"])
        response = {
            "request": request,
            "result": self.plugin.count_ctx_in_date(year, month, day),
        }
        return response

    def prepare_request(self, item) -> dict:
        """
        Prepare request item for result

        :param item: item with parameters
        :return: request item
        """
        return {"cmd": item["cmd"]}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.14 05:00:00                  #
# ================================================== #

class ContextDebug:
    def __init__(self, window=None):
        """
        Context debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'context'

    def update(self):
        """Update debug window"""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, 'current (id)', str(self.window.core.ctx.current))
        self.window.core.debug.add(self.id, 'len(meta)', len(self.window.core.ctx.meta))
        self.window.core.debug.add(self.id, 'len(items)', len(self.window.core.ctx.items))
        if self.window.core.ctx.tmp_meta is not None:
            self.window.core.debug.add(self.id, 'tmp meta', str(self.window.core.ctx.tmp_meta.to_dict()))
        self.window.core.debug.add(self.id, 'group_id (active)', str(self.window.controller.ctx.group_id))
        self.window.core.debug.add(self.id, 'assistant', str(self.window.core.ctx.assistant))
        self.window.core.debug.add(self.id, 'mode', str(self.window.core.ctx.mode))
        self.window.core.debug.add(self.id, 'model', str(self.window.core.ctx.model))
        self.window.core.debug.add(self.id, 'preset', str(self.window.core.ctx.preset))
        self.window.core.debug.add(self.id, 'run', str(self.window.core.ctx.run))
        self.window.core.debug.add(self.id, 'status', str(self.window.core.ctx.status))
        self.window.core.debug.add(self.id, 'thread', str(self.window.core.ctx.thread))
        self.window.core.debug.add(self.id, 'last_mode', str(self.window.core.ctx.last_mode))
        self.window.core.debug.add(self.id, 'last_model', str(self.window.core.ctx.last_model))
        self.window.core.debug.add(self.id, 'search_string', str(self.window.core.ctx.search_string))
        self.window.core.debug.add(self.id, 'filters', str(self.window.core.ctx.filters))
        self.window.core.debug.add(self.id, 'filters_labels', str(self.window.core.ctx.filters_labels))
        self.window.core.debug.add(self.id, 'sys_prompt (current)', str(self.window.core.ctx.current_sys_prompt))
        self.window.core.debug.add(self.id, 'allowed_modes', str(self.window.core.ctx.allowed_modes))
        self.window.core.debug.add(self.id, 'CMD (current)', str(self.window.core.ctx.current_cmd))
        self.window.core.debug.add(self.id, 'CMD schema (current)', str(self.window.core.ctx.current_cmd_schema))

        current = None
        if self.window.core.ctx.current is not None:
            if self.window.core.ctx.current in self.window.core.ctx.meta:
                current = self.window.core.ctx.meta[self.window.core.ctx.current]
            if current is not None:
                data = current.to_dict()
                self.window.core.debug.add(self.id, '*** (current)', str(data))

        i = 0
        self.window.core.debug.add(self.id, 'items[]', '')
        for item in self.window.core.ctx.items:
            data = item.to_dict()
            self.window.core.debug.add(self.id, str(item.id), str(data))
            i += 1

        self.window.core.debug.end(self.id)
